"""Commands to finish up after a demultiplex run."""
import logging

import click

from cg.constants.constants import DRY_RUN
from cg.constants.demultiplexing import OPTION_BCL_CONVERTER, BclConverter
from cg.meta.demultiplex.demux_post_processing import (
    DemuxPostProcessingNovaseqAPI,
    DemuxPostProcessingHiseqXAPI,
)
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.group(name="finish")
def finish_group():
    """Finish up after demultiplexing."""


@finish_group.command(name="all")
@OPTION_BCL_CONVERTER
@DRY_RUN
@click.pass_obj
def finish_all_cmd(context: CGConfig, bcl_converter: str, dry_run: bool) -> None:
    """Command to post-process all demultiplexed flow cells."""
    demux_post_processing_api: DemuxPostProcessingNovaseqAPI = DemuxPostProcessingNovaseqAPI(
        config=context
    )
    demux_post_processing_api.set_dry_run(dry_run=dry_run)
    demux_post_processing_api.finish_all_flow_cells(bcl_converter=bcl_converter)


@finish_group.command(name="flow-cell")
@click.argument("flow-cell-name")
@OPTION_BCL_CONVERTER
@DRY_RUN
@click.option("--force", is_flag=True)
@click.pass_obj
def finish_flow_cell(
    context: CGConfig, flow_cell_name: str, bcl_converter: str, dry_run: bool, force: bool
) -> None:
    """Command to finish up a flow cell after demultiplexing.

    flow-cell-name is full flow cell name, e.g. '201203_A00689_0200_AHVKJCDRXX'.
    """

    demux_post_processing_api: DemuxPostProcessingNovaseqAPI = DemuxPostProcessingNovaseqAPI(
        config=context
    )
    demux_post_processing_api.set_dry_run(dry_run)
    demux_post_processing_api.finish_flow_cell(
        flow_cell_name=flow_cell_name, force=force, bcl_converter=bcl_converter
    )


@finish_group.command(name="all-hiseq-x")
@DRY_RUN
@click.pass_obj
def finish_all_hiseq_x(context: CGConfig, dry_run: bool) -> None:
    """Command to post-process new demultiplexed Hiseq X flow cells."""
    logging.debug("Checking for new Hiseq X demultiplexed flow cells")
    demux_post_processing_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=context
    )
    demux_post_processing_api.set_dry_run(dry_run=dry_run)
    demux_post_processing_api.finish_all_flow_cells(bcl_converter=BclConverter.BCL2FASTQ.value)



from cg.store.filters.status_application_filters import (
    filter_applications_by_entry_id,
    filter_applications_by_prep_category,
    filter_applications_by_tag,
    filter_applications_is_archived,
    filter_applications_is_external,
    filter_applications_is_not_external,
    filter_applications_is_not_archived,
)

from cg.store import Store
from cg.store.models import Application
from sqlalchemy.orm import Query
from tests.store.conftest import StoreConftestFixture


def test_filter_get_application_by_tag(
    store_with_an_application_with_and_without_attributes: Store,
    tag=StoreConftestFixture.TAG_APPLICATION_WITH_ATTRIBUTES.value,
) -> None:
    """Test to get application by tag."""
    # GIVEN a store with two applications of which one has a tag

    # WHEN getting an application by tag
    application: Query = filter_applications_by_tag(
        applications=store_with_an_application_with_and_without_attributes._get_query(
            table=Application
        ),
        tag=tag,
    )

    # ASSERT that application is a query
    assert isinstance(application, Query)

    # THEN assert the application was found
    assert application.all() and len(application.all()) == 1 and application.all()[0].tag == tag


def test_filter_get_applications_by_prep_category(
    store_with_an_application_with_and_without_attributes: Store,
    prep_category=StoreConftestFixture.PREP_CATEGORY_APPLICATION_WITH_ATTRIBUTES.value,
) -> None:
    """Test to get application by prep category."""
    #  GIVEN a store with two applications of which one is of a prep category

    # WHEN getting an application by prep category
    application: Query = filter_applications_by_prep_category(
        applications=store_with_an_application_with_and_without_attributes._get_query(
            table=Application
        ),
        prep_category=prep_category,
    )

    # ASSERT that application is a query
    assert isinstance(application, Query)

    # THEN assert the application was found
    assert (
        application.all()
        and len(application.all()) == 1
        and application.all()[0].prep_category == prep_category
    )


def test_filter_get_applications_is_archived(
    store_with_an_application_with_and_without_attributes: Store,
) -> None:
    """Test to get application by is_archived."""
    # GIVEN a store with two applications of which one is archived

    # WHEN getting an application by is_archived
    application: Query = filter_applications_is_archived(
        applications=store_with_an_application_with_and_without_attributes._get_query(
            table=Application
        )
    )

    # ASSERT that application is a query
    assert isinstance(application, Query)

    # THEN assert the application was found
    assert (
        application.all()
        and len(application.all()) == 1
        and application.all()[0].is_archived is True
    )


def test_filter_application_is_not_archived(
    store_with_an_application_with_and_without_attributes: Store,
) -> None:
    """Test to get application when no archived."""
    # GIVEN a store with two applications of which one is archived

    # WHEN getting an application that is not archived
    application: Query = filter_applications_is_not_archived(
        applications=store_with_an_application_with_and_without_attributes._get_query(
            table=Application
        )
    )

    # ASSERT that application is a query
    assert isinstance(application, Query)

    # THEN assert the application was found
    assert (
        application.all()
        and len(application.all()) == 1
        and application.all()[0].is_archived is False
    )


def test_filter_get_applications_is_external(
    store_with_an_application_with_and_without_attributes: Store,
) -> None:
    """Test to get application by is_external."""
    # GIVEN a store with two applications of which one is external

    # WHEN getting an application by is_external
    application: Query = filter_applications_is_external(
        applications=store_with_an_application_with_and_without_attributes._get_query(
            table=Application
        )
    )

    # ASSERT that application is a query
    assert isinstance(application, Query)

    # THEN assert the application was found
    assert (
        application.all()
        and len(application.all()) == 1
        and application.all()[0].is_external is True
    )


def test_filter_get_applications_is_not_external(
    store_with_an_application_with_and_without_attributes: Store,
) -> None:
    """Test to get application by is_external."""
    # GIVEN a store with two applications of which one is external

    # WHEN getting an application by is_external
    application: Query = filter_applications_is_not_external(
        applications=store_with_an_application_with_and_without_attributes._get_query(
            table=Application
        )
    )

    # ASSERT that application is a query
    assert isinstance(application, Query)

    # THEN assert the application was found
    assert (
        application.all()
        and len(application.all()) == 1
        and application.all()[0].is_external is False
    )


def test_filter_get_applications_by_entry_id(
    store_with_an_application_with_and_without_attributes: Store,
    entry_id: int = 1,
) -> None:
    """Test to get application by id."""
    # GIVEN a database with an application two applications

    # WHEN getting an application by id
    application: Query = filter_applications_by_entry_id(
        applications=store_with_an_application_with_and_without_attributes._get_query(
            table=Application
        ),
        entry_id=entry_id,
    )

    # ASSERT that applications is a query
    assert isinstance(application, Query)

    # THEN assert the application was found
    assert application.all() and len(application.all()) == 1 and application.all()[0].id == entry_id



import logging
from datetime import datetime
from pathlib import Path
from typing import Union, Optional

from cg.apps.coverage import ChanjoAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.meta.workflow.analysis import AnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store import Store

LOG = logging.getLogger(__name__)


class MockMipDNAAnalysisAPI(MipDNAAnalysisAPI):
    """Mock MipDNAAnalysisAPI for CLI tests"""

    def __init__(self, config: CGConfig):
        super().__init__(config)

    @property
    def root(self):
        """docstring for root"""

        return "/root/path"


class MockHousekeeperMipDNAReportAPI(MipDNAReportAPI):
    """Mock ReportAPI for CLI tests overwriting HK methods"""

    def __init__(self, config: CGConfig, analysis_api: AnalysisAPI):
        super().__init__(config, analysis_api)

    def add_delivery_report_to_hk(
        self, delivery_report_file: Path, case_id: str, analysis_date: datetime
    ):
        """docstring for add_delivery_report_to_hk"""

        LOG.info(
            "add_delivery_report_to_hk called with the following args: delivery_report_file=%s, case=%s, "
            "analysis_date=%s",
            delivery_report_file,
            case_id,
            analysis_date,
        )

    def get_delivery_report_from_hk(self, case_id: str) -> str:
        """docstring for get_delivery_report_from_hk"""

        LOG.info(f"get_delivery_report_from_hk called with the following args: case={case_id}")

        return "/path/to/delivery_report.html"

    def get_scout_uploaded_file_from_hk(self, case_id: str, scout_tag: str) -> Optional[str]:
        """docstring for get_scout_uploaded_file_from_hk"""

        LOG.info(
            "add_delivery_report_to_hk called with the following args: case=%s, scout_tag=%s",
            case_id,
            scout_tag,
        )

        return f"path/to/{scout_tag}"


class MockMipDNAReportAPI(MockHousekeeperMipDNAReportAPI):
    """Mock ReportAPI for CLI tests"""

    def __init__(self, config: CGConfig, analysis_api: AnalysisAPI):
        super().__init__(config, analysis_api)

    def create_delivery_report(self, case_id: str, analysis_date: datetime, force_report: bool):
        """docstring for create_delivery_report"""

        LOG.info(
            "create_delivery_report called with the following args: case=%s, analysis_date=%s, force_report=%s",
            case_id,
            analysis_date,
            force_report,
        )

    def create_delivery_report_file(
        self, case_id: str, file_path: Path, analysis_date: datetime, force_report: bool
    ):
        """docstring for create_delivery_report_file"""

        LOG.info(
            "create_delivery_report_file called with the following args: case=%s, file_path=%s, analysis_date=%s, "
            "force_report=%s",
            case_id,
            file_path,
            analysis_date,
            force_report,
        )

        return file_path


class MockDB(Store):
    """Mock database"""

    def __init__(self, store):
        self.store = store


class MockChanjo(ChanjoAPI):
    """Chanjo mock class"""

    def __init__(self):
        mock_config = {"chanjo": {"config_path": "/mock/path", "binary_path": "/mock/path"}}
        super().__init__(mock_config)

    def sample_coverage(self, sample_id: str, panel_genes: list) -> Union[None, dict]:
        """Calculates sample coverage for a specific panel"""

        sample_coverage = None
        if sample_id == "ADM1":
            sample_coverage = {"mean_coverage": 38.342, "mean_completeness": 99.1}
        elif sample_id == "ADM2":
            sample_coverage = {"mean_coverage": 37.342, "mean_completeness": 97.1}
        elif sample_id == "ADM3":
            sample_coverage = {"mean_coverage": 39.342, "mean_completeness": 98.1}

        return sample_coverage



"""Tests the findbusinessdata part of the Cg store API related to Analysis model."""

from typing import List
from cg.store import Store
from cg.store.models import (
    Analysis,
)
from tests.store_helpers import StoreHelpers


def test_get_analysis_by_case(
    store_with_case_and_analysis: Store, case_id: str = "test_case_internal_id"
):
    """Test that an analysis can be fetched by case."""
    # GIVEN a database with an analysis and case
    case = store_with_case_and_analysis.get_case_by_internal_id(internal_id=case_id)
    assert case
    # WHEN fetching the analysis by case
    analyses: List[Analysis] = store_with_case_and_analysis.get_analyses_by_case_entry_id(
        case_entry_id=case.id
    )
    # THEN one analysis should be returned
    for analysis in analyses:
        assert analysis
        assert analysis.family == case



"""Module that handles deletion of flow cell run directories and their BCL files from
flow_cell_run_dir/<sequencer>."""
import logging
import shutil
from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional

from housekeeper.store.models import Bundle

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import FlowCellStatus
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.symbols import UNDERSCORE
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.utils.date import get_timedelta_from_date
from cg.store import Store
from cg.store.models import Flowcell

FLOW_CELL_DATE_POSITION = 0
FLOW_CELL_IDENTIFIER_POSITION = -1
LOG = logging.getLogger(__name__)
PDC_RETRIEVAL_STATUSES = [FlowCellStatus.PROCESSING, FlowCellStatus.RETRIEVED]


class RunDirFlowCell:
    """Class to check and remove flow cell run directories."""

    def __init__(self, flow_cell_dir: Path, status_db: Store, housekeeper_api: HousekeeperAPI):
        self.flow_cell_dir: Path = flow_cell_dir
        self.status_db: Store = status_db
        self.hk: HousekeeperAPI = housekeeper_api
        self.flow_cell_name: str = self.flow_cell_dir.name
        self.identifier: str = self.flow_cell_name.split(UNDERSCORE)[FLOW_CELL_IDENTIFIER_POSITION]
        self.run_date: str = self.flow_cell_name.split(UNDERSCORE)[FLOW_CELL_DATE_POSITION]
        self.derived_date: datetime = datetime.strptime(self.run_date, "%y%m%d")
        self.id: str = self.identifier[1:]
        self._age: Optional[timedelta] = None
        self._is_retrieved_from_pdc: Optional[bool] = None
        self._exists_in_statusdb: Optional[bool] = None
        self._sequenced_date: Optional[datetime] = None
        self._flow_cell_status: Optional[str] = None
        self.sample_sheet_path: Path = (
            self.flow_cell_dir / DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
        )

    @property
    def age(self) -> int:
        """How long ago a flow cell was sequenced, in days."""
        if self._age is None:
            self._age: int = (get_timedelta_from_date(date=self.sequenced_date)).days
        return self._age

    @property
    def sequenced_date(self) -> datetime:
        """The date on which the flow cell was sequenced."""
        if self._sequenced_date is None:
            flow_cell: Flowcell = self.status_db.get_flow_cell(flow_cell_id=self.id)
            if flow_cell:
                LOG.info(f"Found flow cell {self.id} in Statusdb, getting sequenced date.")
                self._sequenced_date: datetime = flow_cell.sequenced_at
            else:
                LOG.info(
                    f"Flow cell {self.id} NOT found in Statusdb, deriving sequenced date from run dir name!"
                )
                self._sequenced_date: datetime = self.derived_date
        return self._sequenced_date

    @property
    def is_retrieved_from_pdc(self) -> bool:
        """Flow cell is (being) retrieved from PDC."""
        if self._is_retrieved_from_pdc is None:
            self._is_retrieved_from_pdc = self.flow_cell_status in PDC_RETRIEVAL_STATUSES
        return self._is_retrieved_from_pdc

    @property
    def flow_cell_status(self) -> str:
        """Return status of the flow cell."""
        if self._flow_cell_status is None:
            self._flow_cell_status = self.status_db.get_flow_cell(flow_cell_id=self.id).status
        return self._flow_cell_status

    @property
    def exists_in_statusdb(self) -> bool:
        """The flow cell exists in Statusdb."""
        if self._exists_in_statusdb is None:
            self._exists_in_statusdb = (
                self.status_db.get_flow_cell(flow_cell_id=self.id) is not None
            )
        return self._exists_in_statusdb

    def remove_run_directory(self) -> None:
        """Removes the flow cell run directory."""
        LOG.info(f"Removing run directory {self.flow_cell_dir}")
        shutil.rmtree(self.flow_cell_dir, ignore_errors=True)

    def archive_sample_sheet(self) -> None:
        """Archives a sample sheet in Housekeeper bundles."""
        LOG.info(f"Start archiving sample sheet {self.sample_sheet_path}")
        if not self.sample_sheet_path.exists():
            LOG.warning("Sample sheet does not exists!")
            return
        LOG.info("Sample sheet found!")
        hk_bundle: Bundle = self.hk.bundle(name=self.id)
        if hk_bundle is None:
            LOG.info(f"Creating bundle with name {self.id}")
            self.hk.create_new_bundle_and_version(name=self.id)
        try:
            self.hk.add_and_include_file_to_latest_version(
                bundle_name=self.id,
                file=self.sample_sheet_path,
                tags=[SequencingFileTag.ARCHIVED_SAMPLE_SHEET, self.id],
            )
        except FileExistsError:
            LOG.warning("Sample sheet already included!")



"""Test the Lims api"""
import datetime as dt

from requests.exceptions import HTTPError

from cg.apps.lims.api import LimsAPI


def test_get_received_date(lims_api, mocker):
    """Test to get the received date"""
    # GIVEN a lims api and a mocked sample that returns a received at date
    mocked_sample = mocker.patch("cg.apps.lims.api.Sample")
    mock_instance = mocked_sample.return_value
    date = dt.datetime.today()
    mock_instance.udf = {"Received at": date}

    # WHEN fetching the date
    res = lims_api.get_received_date(123)

    # THEN asserrt the correct date is fetched
    assert res == date


def test_get_received_date_no_sample(lims_api, mocker):
    """Test to get the received date when sample not exists"""
    # GIVEN a lims api and a mocked sample that raises exception when fething date
    mocked_sample = mocker.patch("cg.apps.lims.api.Sample")
    mock_instance = mocked_sample.return_value
    mock_instance.udf.get.side_effect = HTTPError()

    # WHEN fetching the date
    res = lims_api.get_received_date(123)

    # THEN assert that None is returned since a exception was raised
    assert res is None


def test_get_prepared_date(lims_api, mocker):
    """Test to get the prepared date for an existing sample"""
    # GIVEN a lims api and a mocked sample that returns a prepared at date
    mocked_sample = mocker.patch("cg.apps.lims.api.Sample")
    mock_instance = mocked_sample.return_value
    date = dt.datetime.today()
    mock_instance.udf = {"Library Prep Finished": date}

    # WHEN fetching the date
    res = lims_api.get_prepared_date(123)

    # THEN asserrt the correct date is fetched
    assert res == date


def test_get_prepared_date_no_sample(lims_api, mocker):
    """Test to get the prepared date when sample not exists"""
    # GIVEN a lims api and a mocked sample that raises exception when fething date
    mocked_sample = mocker.patch("cg.apps.lims.api.Sample")
    mock_instance = mocked_sample.return_value
    mock_instance.udf.get.side_effect = HTTPError()

    # WHEN fetching the date
    res = lims_api.get_prepared_date(123)

    # THEN assert that None is returned since a exception was raised
    assert res is None


def test_get_delivery_date(lims_api, mocker):
    """Test to get the delivery date for an existing sample"""
    # GIVEN a lims api and a mocked sample that returns a delivery date
    mocked_sample = mocker.patch("cg.apps.lims.api.Sample")
    mock_instance = mocked_sample.return_value
    date = dt.datetime.today()
    mock_instance.udf = {"Delivered at": date}

    # WHEN fetching the date
    res = lims_api.get_delivery_date(123)

    # THEN asserrt the correct date is fetched
    assert res == date


def test_get_delivery_date_no_sample(lims_api, mocker):
    """Test to get the delivery date when sample not exists"""
    # GIVEN a lims api and a mocked sample that raises exception when fething date
    mocked_sample = mocker.patch("cg.apps.lims.api.Sample")
    mock_instance = mocked_sample.return_value
    mock_instance.udf.get.side_effect = HTTPError()

    # WHEN fetching the date
    res = lims_api.get_delivery_date(123)

    # THEN assert that None is returned since a exception was raised
    assert res is None


def test_get_sequenced_date(lims_api, mocker):
    """Test to get the sequenced date"""
    # GIVEN a lims api and a mocked sample that returns a sequenced date
    mocked_sample = mocker.patch("cg.apps.lims.api.Sample")
    mock_instance = mocked_sample.return_value
    date = dt.datetime.today()
    mock_instance.udf = {"Sequencing Finished": date}

    # WHEN fetching the date
    res = lims_api.get_sequenced_date(123)

    # THEN asserrt the correct date is fetched
    assert res == date


def test_get_sequenced_date_no_sample(lims_api, mocker):
    """Test to get the sequenced date when sample not exists"""
    # GIVEN a lims api and a mocked sample that raises exception when fething date
    mocked_sample = mocker.patch("cg.apps.lims.api.Sample")
    mock_instance = mocked_sample.return_value
    mock_instance.udf.get.side_effect = HTTPError()

    # WHEN fetching the date
    res = lims_api.get_sequenced_date(123)

    # THEN assert that None is returned since a exception was raised
    assert res is None


def test_get_processing_time_for_values(lims_api, mocker):
    """Test to get the processing time when there are some existing dates"""
    # GIVEN there is a received_at date and a delivered date
    date = dt.datetime.today()
    # GIVEN that received and delivered date is the same
    mocker.patch.object(LimsAPI, "get_received_date")
    LimsAPI.get_received_date.return_value = date

    mocker.patch.object(LimsAPI, "get_delivery_date")
    LimsAPI.get_delivery_date.return_value = date

    # WHEN fetching the processing time
    processing_time = lims_api.get_processing_time(123)

    # THEN return value is the delivered - received == 0 for same datetime
    assert processing_time == dt.timedelta(0)


def test_get_processing_time_for_no_received_time(lims_api, mocker):
    """Test to get the processing time when there is not received time"""
    # GIVEN there is no received_at date
    mocker.patch.object(LimsAPI, "get_received_date")
    LimsAPI.get_received_date.return_value = None
    # GIVEN there is a delivery date
    date = dt.datetime.today()
    mocker.patch.object(LimsAPI, "get_delivery_date")
    LimsAPI.get_delivery_date.return_value = date

    # WHEN calling get_processing_time
    processing_time = lims_api.get_processing_time(123)

    # THEN return value should be none since one time is missing
    assert processing_time is None


def test_get_processing_time_for_no_delivered_time(lims_api, mocker):
    """Test to get the processing time when there is not delivered time"""

    # GIVEN there is a received_at date and no delivered date
    date = dt.datetime.today()
    mocker.patch.object(LimsAPI, "get_received_date")
    LimsAPI.get_received_date.return_value = date

    mocker.patch.object(LimsAPI, "get_delivery_date")
    LimsAPI.get_delivery_date.return_value = None

    # WHEN calling get_processing_time
    processing_time = lims_api.get_processing_time(123)

    # THEN return value is none
    assert processing_time is None



"""Priority specific constants"""
from enum import IntEnum

from cg.utils.enums import StrEnum


class SlurmQos(StrEnum):
    MAINTENANCE: str = "maintenance"
    LOW: str = "low"
    NORMAL: str = "normal"
    HIGH: str = "high"
    EXPRESS: str = "express"


class PriorityTerms(StrEnum):
    EXPRESS: str = "express"
    PRIORITY: str = "priority"
    RESEARCH: str = "research"
    STANDARD: str = "standard"
    CLINICAL_TRIALS: str = "clinical_trials"


SLURM_ACCOUNT_TO_QOS = {
    "production": SlurmQos.NORMAL,
    "development": SlurmQos.LOW,
}


class Priority(IntEnum):
    research: int = 0
    standard: int = 1
    priority: int = 2
    express: int = 3
    clinical_trials: int = 4


PRIORITY_TO_SLURM_QOS = {
    Priority.research: SlurmQos.LOW,
    Priority.standard: SlurmQos.NORMAL,
    Priority.priority: SlurmQos.HIGH,
    Priority.express: SlurmQos.EXPRESS,
    Priority.clinical_trials: SlurmQos.NORMAL,
}



"""This script tests the cli method to clean old balsamic run dirs"""
import datetime as dt
import logging
from pathlib import Path

from cg.apps.tb import TrailblazerAPI
from cg.cli.workflow.commands import clean_run_dir, past_run_dirs
from cg.constants import Pipeline
from click.testing import CliRunner

from cg.models.cg_config import CGConfig
from tests.store_helpers import StoreHelpers

EXIT_SUCCESS = 0


def test_past_run_dirs_without_options(cli_runner: CliRunner, clean_context: CGConfig):
    """Test command without options"""
    # GIVEN no case_id or options
    # WHEN dry running without anything specified
    result = cli_runner.invoke(past_run_dirs, obj=clean_context)
    # THEN command should mention argument
    assert result.exit_code != EXIT_SUCCESS
    assert "Missing argument" in result.output


def test_run_dir_without_options(cli_runner: CliRunner, clean_context: CGConfig):
    """Test command without options"""
    # GIVEN no case_id or options
    # WHEN dry running without anything specified
    result = cli_runner.invoke(clean_run_dir, obj=clean_context)
    # THEN command should mention argument
    assert result.exit_code != EXIT_SUCCESS
    assert "Missing argument" in result.output


def test_with_yes(
    balsamic_case_clean: str,
    cli_runner: CliRunner,
    clean_context: CGConfig,
    timestamp_now: dt.datetime,
    helpers: StoreHelpers,
    caplog,
    mocker,
):
    """Test command with dry run options."""
    # GIVEN a case on disk that could be deleted
    analysis_api = clean_context.meta_apis["analysis_api"]

    analysis_to_clean = analysis_api.status_db.get_case_by_internal_id(
        balsamic_case_clean
    ).analyses[0]
    case_path = analysis_api.get_case_path(balsamic_case_clean)
    Path(case_path).mkdir(exist_ok=True, parents=True)

    mocker.patch.object(TrailblazerAPI, "is_latest_analysis_ongoing")
    TrailblazerAPI.is_latest_analysis_ongoing.return_value = False

    # WHEN running with yes and remove stuff from before today
    result = cli_runner.invoke(past_run_dirs, ["-y", str(timestamp_now)], obj=clean_context)
    # THEN the analysis should have been cleaned
    assert result.exit_code == EXIT_SUCCESS
    assert analysis_to_clean.cleaned_at
    assert analysis_to_clean not in analysis_api.get_analyses_to_clean(before=timestamp_now)
    assert not Path(case_path).exists()


def test_dry_run(
    balsamic_case_clean: str,
    cli_runner: CliRunner,
    clean_context: CGConfig,
    timestamp_yesterday: dt.datetime,
    helpers: StoreHelpers,
    caplog,
    mocker,
):
    """Test command with dry run options"""

    # GIVEN a case on disk that could be deleted
    caplog.set_level(logging.INFO)
    base_store = clean_context.status_db
    helpers.add_analysis(
        base_store,
        pipeline=Pipeline.BALSAMIC,
        started_at=timestamp_yesterday,
        uploaded_at=timestamp_yesterday,
        cleaned_at=None,
    )

    analysis_to_clean = base_store.get_case_by_internal_id(balsamic_case_clean).analyses[0]
    case_path = clean_context.meta_apis["analysis_api"].get_case_path(balsamic_case_clean)
    Path(case_path).mkdir(exist_ok=True, parents=True)

    mocker.patch.object(TrailblazerAPI, "is_latest_analysis_ongoing")
    TrailblazerAPI.is_latest_analysis_ongoing.return_value = False

    # WHEN dry running with dry run specified
    result = cli_runner.invoke(clean_run_dir, [balsamic_case_clean, "-d", "-y"], obj=clean_context)
    # THEN command should say it would have deleted
    assert result.exit_code == EXIT_SUCCESS
    assert "Would have deleted" in caplog.text
    assert balsamic_case_clean in caplog.text
    assert analysis_to_clean in base_store.analyses_to_clean(pipeline=Pipeline.BALSAMIC)


def test_cleaned_at_valid(
    balsamic_case_clean: str, cli_runner: CliRunner, clean_context: CGConfig, caplog, mocker
):
    """Test command with dry run options"""
    # GIVEN a case on disk that could be deleted
    base_store = clean_context.status_db
    case_path = clean_context.meta_apis["analysis_api"].get_case_path(balsamic_case_clean)
    Path(case_path).mkdir(exist_ok=True, parents=True)

    mocker.patch.object(TrailblazerAPI, "is_latest_analysis_ongoing")
    TrailblazerAPI.is_latest_analysis_ongoing.return_value = False

    # WHEN dry running with dry run specified
    result = cli_runner.invoke(clean_run_dir, [balsamic_case_clean, "-y"], obj=clean_context)

    # THEN command should say it would have deleted
    assert result.exit_code == EXIT_SUCCESS
    assert base_store.get_case_by_internal_id("balsamic_case_clean").analyses[0].cleaned_at
    assert not Path(case_path).exists()


def test_cleaned_at_invalid(
    balsamic_case_not_clean: str, cli_runner: CliRunner, clean_context: CGConfig, caplog
):
    """Test command with dry run options"""
    # GIVEN a case on disk that could be deleted
    base_store = clean_context.status_db
    case_path = clean_context.meta_apis["analysis_api"].get_case_path(balsamic_case_not_clean)
    Path(case_path).mkdir(exist_ok=True, parents=True)
    assert not base_store.get_case_by_internal_id(balsamic_case_not_clean).analyses[0].cleaned_at
    # WHEN dry running with dry run specified

    result = cli_runner.invoke(past_run_dirs, ["2020-12-01", "-d", "-y"], obj=clean_context)

    # THEN case directory should not have been cleaned
    assert result.exit_code == EXIT_SUCCESS
    assert not base_store.get_case_by_internal_id(balsamic_case_not_clean).analyses[0].cleaned_at
    assert Path(case_path).exists()



from typing import Optional, List, Callable
from enum import Enum
from sqlalchemy.orm import Query

from cg.store.models import Family, Sample


def get_samples_associated_with_case(case_samples: Query, case_id: str, **kwargs) -> Query:
    """Return samples associated with a case."""
    return case_samples.filter(Family.internal_id == case_id)


def get_cases_associated_with_sample(case_samples: Query, sample_id: str, **kwargs) -> Query:
    """Return cases associated with a sample."""
    return case_samples.filter(Sample.internal_id == sample_id)


def get_cases_associated_with_sample_by_entry_id(
    case_samples: Query, sample_entry_id: int, **kwargs
) -> Query:
    """Return cases associated with a sample."""
    return case_samples.filter(Sample.id == sample_entry_id)


def apply_case_sample_filter(
    filter_functions: List[Callable],
    case_samples: Query,
    case_id: Optional[str] = None,
    sample_entry_id: Optional[int] = None,
    sample_id: Optional[str] = None,
) -> Query:
    """Apply filtering functions to the sample queries and return filtered results."""

    for function in filter_functions:
        case_samples: Query = function(
            case_samples=case_samples,
            case_id=case_id,
            sample_entry_id=sample_entry_id,
            sample_id=sample_id,
        )
    return case_samples


class CaseSampleFilter(Enum):
    """Define CaseSample filter functions."""

    GET_SAMPLES_ASSOCIATED_WITH_CASE: Callable = get_samples_associated_with_case
    GET_CASES_ASSOCIATED_WITH_SAMPLE: Callable = get_cases_associated_with_sample
    GET_CASES_ASSOCIATED_WITH_SAMPLE_BY_ENTRY_ID: Callable = (
        get_cases_associated_with_sample_by_entry_id
    )



import datetime as dt
from unittest.mock import patch

import pytest
from cgmodels.cg.constants import Pipeline

from cg.store.models import Customer, Family
from tests.store_helpers import StoreHelpers

from cg.constants import DataDelivery
from cg.exc import OrderError, TicketCreationError
from cg.meta.orders import OrdersAPI
from cg.meta.orders.mip_dna_submitter import MipDnaSubmitter
from cg.models.orders.order import OrderIn, OrderType
from cg.models.orders.samples import MipDnaSample
from cg.store import Store
from cg.store.models import Pool, Sample

PROCESS_LIMS_FUNCTION_OLD = "cg.meta.orders.api.process_lims"
PROCESS_LIMS_FUNCTION = "cg.meta.orders.lims.process_lims"
SUBMITTERS = [
    "fastq_submitter",
    "metagenome_submitter",
    "microbial_submitter",
    "case_submitter",
    "pool_submitter",
]


def test_too_long_order_name():
    # GIVEN order with more than allowed characters name
    long_name = "A super long order name that is longer than sixty-four characters."
    assert len(long_name) > Sample.order.property.columns[0].type.length

    # WHEN placing it in the pydantic order model
    # THEN an error is raised
    with pytest.raises(ValueError):
        OrderIn(name=long_name, customer="", comment="", samples=[])


@patch("cg.meta.orders.ticket_handler.FormDataRequest.submit", return_value=None)
@pytest.mark.parametrize(
    "order_type",
    [
        OrderType.BALSAMIC,
        OrderType.FASTQ,
        OrderType.FLUFFY,
        OrderType.METAGENOME,
        OrderType.MICROSALT,
        OrderType.MIP_DNA,
        OrderType.MIP_RNA,
        OrderType.RML,
        OrderType.SARS_COV_2,
    ],
)
def test_submit(
    mail_patch,
    all_orders_to_submit: dict,
    base_store: Store,
    monkeypatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    ticket_id: str,
    user_mail: str,
    user_name: str,
):
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    monkeypatch_process_lims(monkeypatch, order_data)

    # GIVEN an order and an empty store
    assert not base_store.get_samples()

    # WHEN submitting the order

    result = orders_api.submit(
        project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
    )

    # THEN the result should contain the ticket number for the order
    for record in result["records"]:
        if isinstance(record, Pool):
            assert record.ticket == ticket_id
        elif isinstance(record, Sample):
            assert record.original_ticket == ticket_id
        elif isinstance(record, Family):
            for link_obj in record.links:
                assert link_obj.sample.original_ticket == ticket_id


def monkeypatch_process_lims(monkeypatch, order_data):
    lims_project_data = {"id": "ADM1234", "date": dt.datetime.now()}
    lims_map = {sample.name: f"ELH123A{index}" for index, sample in enumerate(order_data.samples)}
    for submitter in SUBMITTERS:
        monkeypatch.setattr(
            f"cg.meta.orders.{submitter}.process_lims",
            lambda **kwargs: (lims_project_data, lims_map),
        )


@pytest.mark.parametrize(
    "order_type",
    [OrderType.MIP_DNA, OrderType.MIP_RNA, OrderType.BALSAMIC],
)
def test_submit_ticketexception(
    all_orders_to_submit,
    base_store: Store,
    monkeypatch,
    orders_api: OrdersAPI,
    order_type: OrderType,
    user_mail: str,
    user_name: str,
):
    # GIVEN an order that does not have a name (ticket_nr)
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    order_data.name = "dummy_name"
    with patch("cg.apps.osticket.OsTicket") as os_mock:
        orders_api.ticket_handler.osticket = os_mock.return_value
        orders_api.ticket_handler.osticket.open_ticket.side_effect = TicketCreationError("ERROR")

    # WHEN the order is submitted and a TicketCreationError raised
    # THEN the TicketCreationError is not excepted
    with pytest.raises(TicketCreationError):
        orders_api.submit(
            project=order_type,
            order_in=order_data,
            user_name=user_name,
            user_mail=user_mail,
        )


@pytest.mark.parametrize(
    "order_type",
    [OrderType.MIP_DNA, OrderType.MIP_RNA, OrderType.BALSAMIC],
)
def test_submit_illegal_sample_customer(
    all_orders_to_submit: dict,
    monkeypatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    sample_store: Store,
    ticket_id: str,
    user_mail: str,
    user_name: str,
):
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    monkeypatch_process_lims(monkeypatch, order_data)

    # GIVEN we have an order with a customer that is not in the same customer group as customer
    # that the samples originate from
    new_customer = sample_store.add_customer(
        "customer999",
        "customer 999",
        scout_access=True,
        invoice_address="dummy street",
        invoice_reference="dummy nr",
    )
    sample_store.add_commit(new_customer)
    existing_sample: Sample = sample_store.get_samples()[0]
    existing_sample.customer = new_customer
    sample_store.add_commit(existing_sample)

    for sample in order_data.samples:
        sample.internal_id = existing_sample.internal_id

    # WHEN calling submit
    # THEN an OrderError should be raised on illegal customer
    with pytest.raises(OrderError):
        orders_api.submit(
            project=order_type,
            order_in=order_data,
            user_name=user_name,
            user_mail=user_mail,
        )


@patch("cg.meta.orders.ticket_handler.FormDataRequest.submit", return_value=None)
@pytest.mark.parametrize(
    "order_type",
    [OrderType.MIP_DNA, OrderType.MIP_RNA, OrderType.BALSAMIC],
)
def test_submit_scout_legal_sample_customer(
    mail_patch,
    all_orders_to_submit: dict,
    monkeypatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    sample_store: Store,
    user_mail: str,
    user_name: str,
):
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    monkeypatch_process_lims(monkeypatch, order_data)
    # GIVEN we have an order with a customer that is in the same customer group as customer
    # that the samples originate from
    collaboration = sample_store.add_collaboration("customer999only", "customer 999 only group")
    sample_store.add_commit(collaboration)
    sample_customer = sample_store.add_customer(
        "customer1",
        "customer 1",
        scout_access=True,
        invoice_address="dummy street 1",
        invoice_reference="dummy nr",
    )
    order_customer = sample_store.add_customer(
        "customer2",
        "customer 2",
        scout_access=True,
        invoice_address="dummy street 2",
        invoice_reference="dummy nr",
    )
    sample_customer.collaborations.append(collaboration)
    order_customer.collaborations.append(collaboration)
    sample_store.add_commit(sample_customer)
    sample_store.add_commit(order_customer)
    existing_sample: Sample = sample_store.get_samples()[0]
    existing_sample.customer = sample_customer
    sample_store.commit()
    order_data.customer = order_customer.internal_id

    for sample in order_data.samples:
        sample.internal_id = existing_sample.internal_id
        break

    # WHEN calling submit
    # THEN an OrderError should not be raised on illegal customer
    orders_api.submit(
        project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
    )


@pytest.mark.parametrize(
    "order_type",
    [OrderType.MIP_DNA, OrderType.MIP_RNA, OrderType.BALSAMIC],
)
def test_submit_duplicate_sample_case_name(
    all_orders_to_submit: dict,
    monkeypatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    ticket_id: str,
    user_mail: str,
    user_name: str,
):
    # GIVEN we have an order with a case that is already in the database
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    store = orders_api.status
    customer: Customer = store.get_customer_by_customer_id(customer_id=order_data.customer)

    for sample in order_data.samples:
        case_id = sample.family_name
        if not store.get_case_by_name_and_customer(customer=customer, case_name=case_id):
            case: Family = store.add_case(
                data_analysis=Pipeline.MIP_DNA,
                data_delivery=DataDelivery.SCOUT,
                name=case_id,
                ticket=ticket_id,
            )
            case.customer = customer
            store.add_commit(case)
        assert store.get_case_by_name_and_customer(customer=customer, case_name=case_id)

    monkeypatch_process_lims(monkeypatch, order_data)

    # WHEN calling submit
    # THEN an OrderError should be raised on duplicate case name
    with pytest.raises(OrderError):
        orders_api.submit(
            project=order_type,
            order_in=order_data,
            user_name=user_name,
            user_mail=user_mail,
        )


@patch("cg.meta.orders.ticket_handler.FormDataRequest.submit", return_value=None)
@pytest.mark.parametrize(
    "order_type",
    [OrderType.FLUFFY],
)
def test_submit_fluffy_duplicate_sample_case_name(
    mail_patch,
    all_orders_to_submit: dict,
    monkeypatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    ticket_id: str,
    user_mail: str,
    user_name: str,
):
    # GIVEN we have an order with a case that is already in the database
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    monkeypatch_process_lims(monkeypatch, order_data)

    orders_api.submit(
        project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
    )

    # WHEN calling submit
    # THEN an OrderError should be raised on duplicate case name
    with pytest.raises(OrderError):
        orders_api.submit(
            project=order_type,
            order_in=order_data,
            user_name=user_name,
            user_mail=user_mail,
        )


@patch("cg.meta.orders.ticket_handler.FormDataRequest.submit", return_value=None)
def test_submit_unique_sample_case_name(
    mail_patch,
    orders_api: OrdersAPI,
    mip_order_to_submit: dict,
    ticket_id: str,
    user_name: str,
    user_mail: str,
    monkeypatch,
):
    # GIVEN we have an order with a case that is not existing in the database
    order_data = OrderIn.parse_obj(obj=mip_order_to_submit, project=OrderType.MIP_DNA)

    store = orders_api.status

    sample: MipDnaSample
    for sample in order_data.samples:
        case_id = sample.family_name
        customer: Customer = store.get_customer_by_customer_id(customer_id=order_data.customer)
        assert not store.get_case_by_name_and_customer(customer=customer, case_name=case_id)

    monkeypatch_process_lims(monkeypatch, order_data)

    # WHEN calling submit
    orders_api.submit(
        project=OrderType.MIP_DNA,
        order_in=order_data,
        user_name=user_name,
        user_mail=user_mail,
    )

    # Then no exception about duplicate names should be thrown


def test_validate_sex_inconsistent_sex(
    orders_api: OrdersAPI, mip_order_to_submit: dict, helpers: StoreHelpers
):
    # GIVEN we have an order with a sample that is already in the database but with different sex
    order_data = OrderIn.parse_obj(mip_order_to_submit, project=OrderType.MIP_DNA)
    store = orders_api.status
    customer: Customer = store.get_customer_by_customer_id(customer_id=order_data.customer)

    # add sample with different sex than in order
    sample: MipDnaSample
    for sample in order_data.samples:
        sample_obj: Sample = helpers.add_sample(
            store=store,
            subject_id=sample.subject_id,
            name=sample.name,
            gender="male" if sample.sex == "female" else "female",
            customer_id=customer.internal_id,
        )
        store.add_commit(sample_obj)
        assert sample_obj.sex != sample.sex

    submitter: MipDnaSubmitter = MipDnaSubmitter(lims=orders_api.lims, status=orders_api.status)

    # WHEN calling _validate_sex
    # THEN an OrderError should be raised on non-matching sex
    with pytest.raises(OrderError):
        submitter._validate_subject_sex(samples=order_data.samples, customer_id=order_data.customer)


def test_validate_sex_consistent_sex(
    orders_api: OrdersAPI, mip_order_to_submit: dict, helpers: StoreHelpers
):
    # GIVEN we have an order with a sample that is already in the database and with same gender
    order_data = OrderIn.parse_obj(mip_order_to_submit, project=OrderType.MIP_DNA)
    store = orders_api.status
    customer: Customer = store.get_customer_by_customer_id(customer_id=order_data.customer)

    # add sample with different sex than in order
    sample: MipDnaSample
    for sample in order_data.samples:
        sample_obj: Sample = helpers.add_sample(
            store=store,
            subject_id=sample.subject_id,
            name=sample.name,
            gender=sample.sex,
            customer_id=customer.internal_id,
        )
        store.add_commit(sample_obj)
        assert sample_obj.sex == sample.sex

    submitter: MipDnaSubmitter = MipDnaSubmitter(lims=orders_api.lims, status=orders_api.status)

    # WHEN calling _validate_sex
    submitter._validate_subject_sex(samples=order_data.samples, customer_id=order_data.customer)

    # THEN no OrderError should be raised on non-matching sex


def test_validate_sex_unknown_existing_sex(
    orders_api: OrdersAPI, mip_order_to_submit: dict, helpers: StoreHelpers
):
    # GIVEN we have an order with a sample that is already in the database and with different gender but the existing is
    # of type "unknown"
    order_data = OrderIn.parse_obj(mip_order_to_submit, project=OrderType.MIP_DNA)
    store = orders_api.status
    customer: Customer = store.get_customer_by_customer_id(customer_id=order_data.customer)

    # add sample with different sex than in order
    sample: MipDnaSample
    for sample in order_data.samples:
        sample_obj: Sample = helpers.add_sample(
            store=store,
            subject_id=sample.subject_id,
            name=sample.name,
            gender="unknown",
            customer_id=customer.internal_id,
        )
        store.add_commit(sample_obj)
        assert sample_obj.sex != sample.sex

    submitter: MipDnaSubmitter = MipDnaSubmitter(lims=orders_api.lims, status=orders_api.status)

    # WHEN calling _validate_sex
    submitter._validate_subject_sex(samples=order_data.samples, customer_id=order_data.customer)

    # THEN no OrderError should be raised on non-matching sex


def test_validate_sex_unknown_new_sex(
    orders_api: OrdersAPI, mip_order_to_submit: dict, helpers: StoreHelpers
):
    # GIVEN we have an order with a sample that is already in the database and with different gender but the new is of
    # type "unknown"
    order_data = OrderIn.parse_obj(mip_order_to_submit, project=OrderType.MIP_DNA)
    store = orders_api.status
    customer: Customer = store.get_customer_by_customer_id(customer_id=order_data.customer)

    # add sample with different sex than in order
    for sample in order_data.samples:
        sample_obj: Sample = helpers.add_sample(
            store=store,
            subject_id=sample.subject_id,
            name=sample.name,
            gender=sample.sex,
            customer_id=customer.internal_id,
        )
        sample.sex = "unknown"
        store.add_commit(sample_obj)

    for sample in order_data.samples:
        assert sample_obj.sex != sample.sex

    submitter: MipDnaSubmitter = MipDnaSubmitter(lims=orders_api.lims, status=orders_api.status)

    # WHEN calling _validate_sex
    submitter._validate_subject_sex(samples=order_data.samples, customer_id=order_data.customer)

    # THEN no OrderError should be raised on non-matching sex


@patch("cg.meta.orders.ticket_handler.FormDataRequest.submit", return_value=None)
@pytest.mark.parametrize(
    "order_type",
    [
        OrderType.BALSAMIC,
        OrderType.FASTQ,
        OrderType.FLUFFY,
        OrderType.METAGENOME,
        OrderType.MICROSALT,
        OrderType.MIP_DNA,
        OrderType.MIP_RNA,
        OrderType.RML,
        OrderType.SARS_COV_2,
    ],
)
def test_submit_unique_sample_name(
    mail_patch,
    all_orders_to_submit: dict,
    monkeypatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    ticket_id: str,
    user_mail: str,
    user_name: str,
):
    # GIVEN we have an order with a sample that is not existing in the database
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    store = orders_api.status
    assert not store.get_samples()

    monkeypatch_process_lims(monkeypatch, order_data)

    # WHEN calling submit
    orders_api.submit(
        project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
    )

    # Then no exception about duplicate names should be thrown


@pytest.mark.parametrize(
    "order_type",
    [OrderType.SARS_COV_2, OrderType.METAGENOME],
)
def test_sarscov2_submit_duplicate_sample_name(
    all_orders_to_submit: dict,
    helpers: StoreHelpers,
    monkeypatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    sample_store: Store,
    user_mail: str,
    user_name: str,
):
    # GIVEN we have an order with samples that is already in the database
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    monkeypatch_process_lims(monkeypatch, order_data)
    store_samples_with_names_from_order(orders_api.status, helpers, order_data)

    # WHEN calling submit
    # THEN an OrderError should be raised on duplicate sample name
    with pytest.raises(OrderError):
        orders_api.submit(
            project=order_type,
            order_in=order_data,
            user_name=user_name,
            user_mail=user_mail,
        )


def store_samples_with_names_from_order(store: Store, helpers: StoreHelpers, order_data: OrderIn):
    customer: Customer = store.get_customer_by_customer_id(customer_id=order_data.customer)
    for sample in order_data.samples:
        sample_name = sample.name
        if not store.get_sample_by_customer_and_name(
            customer_entry_id=[customer.id], sample_name=sample_name
        ):
            sample_obj = helpers.add_sample(
                store=store, name=sample_name, customer_id=customer.internal_id
            )
            store.add_commit(sample_obj)


@patch("cg.meta.orders.ticket_handler.FormDataRequest.submit", return_value=None)
@pytest.mark.parametrize(
    "order_type",
    [
        OrderType.BALSAMIC,
        OrderType.FASTQ,
        OrderType.MICROSALT,
        OrderType.MIP_DNA,
        OrderType.MIP_RNA,
        OrderType.RML,
    ],
)
def test_not_sarscov2_submit_duplicate_sample_name(
    mail_patch,
    all_orders_to_submit: dict,
    helpers: StoreHelpers,
    monkeypatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    sample_store: Store,
    user_mail: str,
    user_name: str,
):
    # GIVEN we have an order with samples that is already in the database
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    monkeypatch_process_lims(monkeypatch, order_data)
    store_samples_with_names_from_order(orders_api.status, helpers, order_data)

    # WHEN calling submit
    orders_api.submit(
        project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
    )

    # THEN no OrderError should be raised on duplicate sample name



"""Delivery report helpers"""

import logging
from datetime import datetime
from typing import Optional

import click
from cgmodels.cg.constants import Pipeline

from cg.constants import REPORT_SUPPORTED_PIPELINES, REPORT_SUPPORTED_DATA_DELIVERY
from cg.meta.report.report_api import ReportAPI
from cg.meta.report.balsamic import BalsamicReportAPI
from cg.meta.report.balsamic_umi import BalsamicUmiReportAPI
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.store.models import Family

LOG = logging.getLogger(__name__)


def get_report_case(context: click.Context, case_id: str) -> Family:
    """Extracts a case object for delivery report generation."""

    # Default report API (MIP DNA report API)
    report_api: ReportAPI = (
        context.obj.meta_apis.get("report_api")
        if context.obj.meta_apis.get("report_api")
        else MipDNAReportAPI(config=context.obj, analysis_api=MipDNAAnalysisAPI(config=context.obj))
    )

    case: Family = report_api.status_db.get_case_by_internal_id(internal_id=case_id)

    # Missing or not valid internal case ID
    if not case_id or not case:
        LOG.warning("Invalid case ID. Retrieving available cases.")

        pipeline = (
            report_api.analysis_api.pipeline if context.obj.meta_apis.get("report_api") else None
        )

        cases_without_delivery_report = (
            report_api.get_cases_without_delivery_report(pipeline)
            if not context.obj.meta_apis.get("upload_api")
            else report_api.get_cases_without_uploaded_delivery_report(pipeline)
        )

        if not cases_without_delivery_report:
            click.echo(
                click.style(
                    "There are no valid cases to perform delivery report actions", fg="green"
                )
            )
        else:
            LOG.error("Provide one of the following case IDs:")
            for case in cases_without_delivery_report:
                click.echo(f"{case.internal_id} ({ case.data_analysis})")

        raise click.Abort

    if case.data_analysis not in REPORT_SUPPORTED_PIPELINES:
        LOG.error(
            f"The {case.data_analysis} pipeline does not support delivery reports (case: {case.internal_id})"
        )

    if case.data_delivery not in REPORT_SUPPORTED_DATA_DELIVERY:
        LOG.error(
            f"The {case.data_delivery} data delivery does not support delivery reports (case: {case.internal_id})"
        )

        raise click.Abort

    return case


def get_report_api(context: click.Context, case: Family) -> ReportAPI:
    """Returns a report API to be used for the delivery report generation."""

    if context.obj.meta_apis.get("report_api"):
        return context.obj.meta_apis.get("report_api")

    return get_report_api_pipeline(context, case.data_analysis)


def get_report_api_pipeline(context: click.Context, pipeline: Pipeline) -> ReportAPI:
    """Resolves the report API given a specific pipeline."""

    # Default report API pipeline: MIP-DNA
    pipeline = pipeline if pipeline else Pipeline.MIP_DNA

    dispatch_report_api = {
        Pipeline.MIP_DNA: MipDNAReportAPI(
            config=context.obj, analysis_api=MipDNAAnalysisAPI(config=context.obj)
        ),
        Pipeline.BALSAMIC: BalsamicReportAPI(
            config=context.obj, analysis_api=BalsamicAnalysisAPI(config=context.obj)
        ),
        Pipeline.BALSAMIC_UMI: BalsamicUmiReportAPI(
            config=context.obj, analysis_api=BalsamicUmiAnalysisAPI(config=context.obj)
        ),
    }

    return dispatch_report_api.get(pipeline)


def get_report_analysis_started(
    case: Family, report_api: ReportAPI, analysis_started_at: Optional[str]
) -> datetime:
    """Resolves and returns a valid analysis date."""

    if not analysis_started_at:
        analysis_started_at: datetime = (
            report_api.status_db.get_case_by_internal_id(case.internal_id).analyses[0].started_at
        )

    # If there is no analysis for the provided date
    if not report_api.status_db.analysis(case, analysis_started_at):
        LOG.error(f"There is no analysis started at {analysis_started_at}")
        raise click.Abort

    LOG.info("Using analysis started at: %s", analysis_started_at)
    return analysis_started_at



import logging
from typing import Optional, Tuple, List

import click
from cg.constants import STATUS_OPTIONS, DataDelivery, Pipeline
from cg.constants.subject import Gender
from cg.meta.transfer.external_data import ExternalDataAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.utils.click.EnumChoice import EnumChoice
from cg.store.models import (
    Family,
    Sample,
    Customer,
    User,
    Collaboration,
    Customer,
    Application,
    Panel,
)

from cg.constants import Priority

LOG = logging.getLogger(__name__)


@click.group()
def add():
    """Add new things to the database."""
    pass


@add.command()
@click.argument("internal_id")
@click.argument("name")
@click.option(
    "-cg",
    "--collaboration-internal-ids",
    "collaboration_internal_ids",
    multiple=True,
    help="List of internal IDs for the collaborations the customer should belong to",
)
@click.option(
    "-ia",
    "--invoice-address",
    "invoice_address",
    required=True,
    help="Street address, Post code, City",
)
@click.option(
    "-ir",
    "--invoice-reference",
    "invoice_reference",
    required=True,
    help="Invoice reference (text)",
)
@click.pass_obj
def customer(
    context: CGConfig,
    internal_id: str,
    name: str,
    collaboration_internal_ids: Optional[List[str]],
    invoice_address: str,
    invoice_reference: str,
):
    """Add a new customer with a unique internal id and name."""
    collaboration_internal_ids = collaboration_internal_ids or []
    status_db: Store = context.status_db

    existing_customer: Customer = status_db.get_customer_by_customer_id(customer_id=internal_id)
    if existing_customer:
        LOG.error(f"{existing_customer.name}: customer already added")
        raise click.Abort

    collaborations: List[Collaboration] = [
        status_db.get_collaboration_by_internal_id(internal_id=collaboration_internal_id)
        for collaboration_internal_id in collaboration_internal_ids
    ]

    new_customer: Customer = status_db.add_customer(
        internal_id=internal_id,
        name=name,
        invoice_address=invoice_address,
        invoice_reference=invoice_reference,
    )
    for collaboration in collaborations:
        new_customer.collaborations.append(collaboration)
    status_db.add_commit(new_customer)
    message: str = f"customer added: {new_customer.internal_id} ({new_customer.id})"
    LOG.info(message)


@add.command()
@click.option("-a", "--admin", is_flag=True, help="make the user an admin")
@click.option(
    "-c",
    "--customer",
    "customer_id",
    required=True,
    help="internal ID for the customer of the user",
)
@click.argument("email")
@click.argument("name")
@click.pass_obj
def user(context: CGConfig, admin: bool, customer_id: str, email: str, name: str):
    """Add a new user with an EMAIL (login) and a NAME (full)."""
    status_db: Store = context.status_db

    customer_obj: Customer = status_db.get_customer_by_customer_id(customer_id=customer_id)
    existing_user: User = status_db.get_user_by_email(email=email)
    if existing_user:
        LOG.error(f"{existing_user.name}: user already added")
        raise click.Abort

    new_user: User = status_db.add_user(
        customer=customer_obj, email=email, name=name, is_admin=admin
    )
    status_db.add_commit(new_user)
    LOG.info(f"User added: {new_user.email} ({new_user.id})")


@add.command()
@click.option("-l", "--lims", "lims_id", help="LIMS id for the sample")
@click.option(
    "-d", "--down-sampled", type=int, help="How many reads is the sample down sampled to?"
)
@click.option("-o", "--order", help="Name of the order the sample belongs to")
@click.option(
    "-s",
    "--sex",
    type=EnumChoice(Gender, use_value=False),
    required=True,
    help="Sample pedigree sex",
)
@click.option("-a", "--application-tag", required=True, help="application tag name")
@click.option(
    "-p",
    "--priority",
    type=EnumChoice(Priority, use_value=False),
    default="standard",
    help="Set the priority for the samples",
)
@click.argument("customer_id")
@click.argument("name")
@click.pass_obj
def sample(
    context: CGConfig,
    lims_id: Optional[str],
    down_sampled: Optional[int],
    sex: Gender,
    order: Optional[str],
    application_tag: str,
    priority: Priority,
    customer_id: str,
    name: str,
):
    """Add a sample for CUSTOMER_ID with a NAME (display)."""
    status_db: Store = context.status_db

    customer: Customer = status_db.get_customer_by_customer_id(customer_id=customer_id)
    if not customer:
        LOG.error(f"Customer: {customer_id} not found")
        raise click.Abort
    application: Application = status_db.get_application_by_tag(tag=application_tag)
    if not application:
        LOG.error(f"Application: {application_tag} not found")

        raise click.Abort
    new_record: Sample = status_db.add_sample(
        name=name,
        sex=sex,
        downsampled_to=down_sampled,
        internal_id=lims_id,
        order=order,
        priority=priority,
    )
    new_record.application_version = status_db.current_application_version(application_tag)
    new_record.customer = customer
    status_db.add_commit(new_record)
    LOG.info(f"{new_record.internal_id}: new sample added")


@add.command()
@click.option(
    "--priority",
    type=EnumChoice(Priority, use_value=False),
    default="standard",
    help="analysis priority",
)
@click.option("-p", "--panel", "panel_abbreviations", multiple=True, help="default gene panels")
@click.option(
    "-a",
    "--analysis",
    "data_analysis",
    help="Analysis workflow",
    required=True,
    type=EnumChoice(Pipeline),
)
@click.option(
    "-dd",
    "--data-delivery",
    "data_delivery",
    help="Update case data delivery",
    required=True,
    type=EnumChoice(DataDelivery),
)
@click.option("-t", "--ticket", help="Ticket number", required=True)
@click.argument("customer_id")
@click.argument("name")
@click.pass_obj
def case(
    context: CGConfig,
    priority: Priority,
    panel_abbreviations: Tuple[str],
    data_analysis: Pipeline,
    data_delivery: DataDelivery,
    customer_id: str,
    name: str,
    ticket: str,
):
    """Add a case with the given name and associated with the given customer"""
    status_db: Store = context.status_db

    customer: Customer = status_db.get_customer_by_customer_id(customer_id=customer_id)
    if customer is None:
        LOG.error(f"{customer_id}: customer not found")
        raise click.Abort

    for panel_abbreviation in panel_abbreviations:
        panel: Panel = status_db.get_panel_by_abbreviation(abbreviation=panel_abbreviation)

        if panel is None:
            LOG.error(f"{panel_abbreviation}: panel not found")
            raise click.Abort

    new_case: Family = status_db.add_case(
        data_analysis=data_analysis,
        data_delivery=data_delivery,
        name=name,
        panels=list(panel_abbreviations),
        priority=priority,
        ticket=ticket,
    )

    new_case.customer = customer
    status_db.add_commit(new_case)
    LOG.info(f"{new_case.internal_id}: new case added")


@add.command()
@click.option("-m", "--mother-id", help="Sample ID for mother of sample")
@click.option("-f", "--father-id", help="Sample ID for father of sample")
@click.option("-s", "--status", type=click.Choice(STATUS_OPTIONS), required=True)
@click.argument("case-id")
@click.argument("sample-id")
@click.pass_obj
def relationship(
    context: CGConfig,
    mother_id: Optional[str],
    father_id: Optional[str],
    status: str,
    case_id: str,
    sample_id: str,
):
    """Create a link between a case id and a sample id."""
    status_db: Store = context.status_db
    mother: Optional[Sample] = None
    father: Optional[Sample] = None
    case_obj: Family = status_db.get_case_by_internal_id(internal_id=case_id)
    if case_obj is None:
        LOG.error("%s: family not found", case_id)
        raise click.Abort

    sample: Sample = status_db.get_sample_by_internal_id(internal_id=sample_id)
    if sample is None:
        LOG.error("%s: sample not found", sample_id)
        raise click.Abort

    if mother_id:
        mother: Sample = status_db.get_sample_by_internal_id(internal_id=mother_id)
        if mother is None:
            LOG.error("%s: mother not found", mother_id)
            raise click.Abort

    if father_id:
        father: Sample = status_db.get_sample_by_internal_id(internal_id=father_id)
        if father is None:
            LOG.error("%s: father not found", father_id)
            raise click.Abort

    new_record = status_db.relate_sample(
        family=case_obj, sample=sample, status=status, mother=mother, father=father
    )
    status_db.add_commit(new_record)
    LOG.info("related %s to %s", case_obj.internal_id, sample.internal_id)


@add.command()
@click.option(
    "-t",
    "--ticket",
    type=str,
    help="Ticket id",
    required=True,
)
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def external(context: CGConfig, ticket: str, dry_run: bool):
    """Downloads external data from the delivery server and places it in appropriate folder on
    the HPC"""
    external_data_api = ExternalDataAPI(config=context)
    external_data_api.transfer_sample_files_from_source(ticket=ticket, dry_run=dry_run)


@add.command("external-hk")
@click.option(
    "-t",
    "--ticket",
    type=str,
    help="Ticket id",
    required=True,
)
@click.option("--dry-run", is_flag=True)
@click.option(
    "--force", help="Overwrites any any previous samples in the customer directory", is_flag=True
)
@click.pass_obj
def external_hk(context: CGConfig, ticket: str, dry_run: bool, force):
    """Adds external data to Housekeeper"""
    external_data_api = ExternalDataAPI(config=context)
    external_data_api.add_transfer_to_housekeeper(dry_run=dry_run, ticket=ticket, force=force)



"""Tests delivery report models validators"""
from cgmodels.cg.constants import Pipeline

from cg.constants import NA_FIELD, YES_FIELD, REPORT_GENDER
from cg.constants.subject import Gender
from cg.models.orders.constants import OrderType
from cg.models.report.validators import (
    validate_empty_field,
    validate_boolean,
    validate_float,
    validate_list,
    validate_rml_sample,
    validate_supported_pipeline,
    validate_gender,
)


def test_validate_empty_field():
    """Tests formatting an empty value"""

    # GIVEN a not valid empty field
    none_field = None

    # WHEN performing the validation
    output = validate_empty_field(none_field)

    # THEN check if the input value was formatted correctly
    assert output == NA_FIELD


def test_validate_boolean():
    """Tests boolean formatting for the delivery report"""

    # GIVEN a not formatted inputs
    none_field = None
    true_field = True
    not_bool_field = "not a boolean"

    # WHEN performing the validation
    validated_none_field = validate_boolean(none_field)
    validated_true_field = validate_boolean(true_field)
    validated_not_bool_field = validate_boolean(not_bool_field)

    # THEN check if the input values were formatted correctly
    assert validated_none_field == NA_FIELD
    assert validated_true_field == YES_FIELD
    assert validated_not_bool_field == NA_FIELD


def test_validate_float():
    """Tests the validation of a float value"""

    # GIVEN a valid float input (float and string format)
    float_value = 12.3456789
    str_value = "12.3456789"

    # WHEN performing the validation
    validated_float_value = validate_float(float_value)
    validated_str_value = validate_float(str_value)

    # THEN check if the input values were formatted correctly
    assert validated_float_value == "12.35"
    assert validated_str_value == "12.35"


def test_validate_float_zero_input():
    """Tests the validation of a float value"""

    # GIVEN a valid float input (float and string format)
    float_value = 0.0
    str_value = "0.0"

    # WHEN performing the validation
    validated_float_value = validate_float(float_value)
    validated_str_value = validate_float(str_value)

    # THEN check if the input values were formatted correctly
    assert validated_float_value == "0.0"
    assert validated_str_value == "0.0"


def test_validate_list():
    """Tests if a list is transformed into a string of comma separated values"""

    # GIVEN a mock list
    mock_list = ["I'm", "a", "list"]

    # WHEN performing the validation
    validated_list = validate_list(mock_list)

    # THEN check if the input values were formatted correctly
    assert validated_list == "I'm, a, list"


def test_validate_rml_sample(caplog):
    """Performs validation on a preparation category"""

    # GIVEN an invalid prep category
    prep_category = OrderType.RML

    # WHEN performing the validation
    try:
        validate_rml_sample(prep_category)
        assert False
    # THEN check if an exception was raised
    except ValueError:
        assert "The delivery report generation does not support RML samples" in caplog.text


def test_validate_gender(caplog):
    """Tests report gender parsing"""

    # GIVEN an invalid gender category
    gender = Gender.FEMALE
    invalid_gender = "not_a_gender"

    # WHEN performing the validation
    validated_gender = validate_gender(gender)
    validated_invalid_gender = validate_gender(invalid_gender)

    # THEN check if the gender has been correctly formatted
    assert validated_gender == REPORT_GENDER.get("female")
    assert validated_invalid_gender == NA_FIELD


def test_validate_supported_pipeline_match_error(caplog):
    """Tests if a customer requested pipeline matches the data analysis one"""

    # GIVEN an input dictionary where the customers and executed pipeline are different
    dict_different_pipelines = {"customer_pipeline": Pipeline.MIP_DNA, "pipeline": Pipeline.FLUFFY}

    # WHEN performing the validation
    try:
        validate_supported_pipeline(None, dict_different_pipelines)
        assert False
    # THEN check if an exception was raised
    except ValueError:
        assert (
            f"The analysis requested by the customer ({dict_different_pipelines.get('customer_pipeline')}) does not "
            f"match the one executed ({dict_different_pipelines.get('pipeline')})" in caplog.text
        )


def test_validate_supported_pipeline(caplog):
    """Tests that the analysis pipeline is supported by the delivery report workflow"""

    # GIVEN a dictionary with a not supported pipeline
    dict_invalid_pipeline = {"customer_pipeline": Pipeline.FLUFFY, "pipeline": Pipeline.FLUFFY}

    # WHEN performing the validation
    try:
        validate_supported_pipeline(None, dict_invalid_pipeline)
        assert False
    # THEN check if an exception was raised
    except ValueError:
        assert (
            f"The pipeline {dict_invalid_pipeline.get('pipeline')} does not support delivery report generation"
            in caplog.text
        )



from typing import List, Optional

from sqlalchemy.orm import Query

from cg.constants import FlowCellStatus
from cg.store import Store
from cg.store.models import Flowcell, Family, Sample
from cg.store.filters.status_flow_cell_filters import (
    get_flow_cell_by_id,
    get_flow_cells_with_statuses,
    get_flow_cells_by_case,
    get_flow_cell_by_id_and_by_enquiry,
)
from tests.store_helpers import StoreHelpers


def test_get_flow_cells_by_case(
    base_store: Store,
    case_id: str,
    case_obj: Family,
    flow_cell_id: str,
    helpers: StoreHelpers,
    sample_obj: Sample,
):
    """Test that a flow cell is returned when there is a flow cell with matching flow cell case."""

    # GIVEN a flow cell that exist in status db
    helpers.add_flowcell(store=base_store, flow_cell_id=flow_cell_id, samples=[sample_obj])

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell: Optional[List[Flowcell]] = get_flow_cells_by_case(
        flow_cells=base_store._get_join_flow_cell_sample_links_query(), case=case_obj
    )

    # THEN a Flowcell type should be returned
    assert isinstance(returned_flow_cell[0], Flowcell)

    # THEN returned flow cell should have the same name as the flow cell in the database
    assert returned_flow_cell[0].name == flow_cell_id


def test_get_flow_cells_by_case_when_no_flow_cell_for_case(
    base_store: Store,
    case_id: str,
    case_obj: Family,
    flow_cell_id: str,
    helpers: StoreHelpers,
    sample_obj: Sample,
):
    """Test that a flow cell is not returned when there is a flow cell with no matching flow cell for case."""

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell: Optional[List[Flowcell]] = get_flow_cells_by_case(
        flow_cells=base_store._get_join_flow_cell_sample_links_query(), case=case_obj
    )

    # THEN returned flow cell should be the original flow cell
    assert not list(returned_flow_cell)


def test_get_flow_cell_by_id(base_store: Store, helpers: StoreHelpers, flow_cell_id: str):
    """Test that a flow cell is returned when there is a flow cell with matching flow cell id."""

    # GIVEN a flow cell that exist in status db
    flow_cell: Flowcell = helpers.add_flowcell(store=base_store, flow_cell_id=flow_cell_id)

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell: Flowcell = get_flow_cell_by_id(
        flow_cells=base_store._get_query(table=Flowcell), flow_cell_id=flow_cell_id
    )

    # THEN returned flow cell should be the original flow cell
    assert isinstance(returned_flow_cell, Query)

    assert flow_cell is returned_flow_cell.first()


def test_get_flow_cell_by_id_and_by_enquiry(
    base_store: Store, helpers: StoreHelpers, flow_cell_id: str
):
    """Test that a flow cell is returned when there is a flow cell with enquiry flow cell id."""

    # GIVEN a flow cell that exist in status db
    flow_cell: Flowcell = helpers.add_flowcell(store=base_store, flow_cell_id=flow_cell_id)

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell: List[Flowcell] = get_flow_cell_by_id_and_by_enquiry(
        flow_cells=base_store._get_query(table=Flowcell), flow_cell_id=flow_cell_id[:4]
    )

    # THEN a list of flow cells should be returned
    assert isinstance(returned_flow_cell[0], Flowcell)

    assert flow_cell is returned_flow_cell[0]


def test_get_flow_cells_with_statuses(base_store: Store, helpers: StoreHelpers, flow_cell_id: str):
    """Test that a flow cell is returned when there is a flow cell with matching flow cell id."""

    # GIVEN a flow cell that exist in status db
    helpers.add_flowcell(store=base_store, flow_cell_id=flow_cell_id)

    # GIVEN a flow cell Query

    # WHEN getting flow cell
    returned_flow_cell_query: Query = get_flow_cells_with_statuses(
        flow_cells=base_store._get_query(table=Flowcell),
        flow_cell_statuses=[FlowCellStatus.ON_DISK, FlowCellStatus.PROCESSING],
    )

    # THEN a query should be returned
    assert isinstance(returned_flow_cell_query, Query)



"""Test for UploadVogueAPI"""

import mock

from cg.constants.constants import FileFormat
from cg.io.controller import ReadStream
from cg.meta.upload.vogue import UploadVogueAPI
from typing import List, Dict


def test_load_genotype(genotype_api, vogue_api, genotype_return, mocker, store):
    """Test load_genotype"""

    # GIVEN UploadVogueAPI and a genotype_return_sample

    # WHEN running load_genotype
    mocker.patch.object(vogue_api, "load_genotype_data")
    mocker.patch.object(genotype_api, "export_sample")
    mocker.patch.object(genotype_api, "export_sample_analysis")

    genotype_api.export_sample.return_value = genotype_return["sample"]
    genotype_api.export_sample_analysis.return_value = genotype_return["sample_analysis"]
    upload_vogue_api = UploadVogueAPI(genotype_api=genotype_api, vogue_api=vogue_api, store=store)
    upload_vogue_api.load_genotype(days="1")

    # THEN vogueapi.load_genotype will be called once for each sample in genotype_return_value
    samples = ReadStream.get_content_from_stream(
        file_format=FileFormat.JSON, stream=genotype_return["sample"]
    )
    call_list = vogue_api.load_genotype_data.call_args_list

    assert vogue_api.load_genotype_data.call_count == 4
    for call in call_list:
        assert call[0][0]["_id"] in samples.keys()


def test_load_apptags(vogue_api, genotype_api, store, mocker):
    """Test load application tags"""
    # GIVEN UploadVogueAPI and a set of application tags
    application_tags: List[Dict] = store.get_applications().apptag_list

    mocker.patch.object(vogue_api, "load_apptags")
    upload_vogue_api = UploadVogueAPI(genotype_api=genotype_api, vogue_api=vogue_api, store=store)

    # WHEN running load_apptags
    upload_vogue_api.load_apptags()

    # THEN load_apptags is called with the application tags inside upload_vogue_api
    vogue_api.load_apptags.assert_called_with(application_tags)


@mock.patch("cg.store.Store")
@mock.patch("cg.apps.vogue.VogueAPI")
@mock.patch("cg.apps.gt.GenotypeAPI")
@mock.patch("cg.store.models.Analysis")
def test_update_analysis_uploaded_to_vogue_date_given_date(
    mock_analysis, mock_genotype_api, mock_vogue_api, mock_store, uploaded_to_vogue_date
):
    """tests updating the uploaded_to_vogue_at field of a record in the analysis table"""

    # GIVEN an analysis object with no uploaded_to_vogue date

    # WHEN setting the uploaded to vogue date to a specific date
    result = UploadVogueAPI(
        genotype_api=mock_genotype_api, vogue_api=mock_vogue_api, store=mock_store
    ).update_analysis_uploaded_to_vogue_date(mock_analysis, uploaded_to_vogue_date)

    # THEN the analysis object should have a vogue_uploaded_date set to vogue_upload_date
    assert result.uploaded_to_vogue_at == uploaded_to_vogue_date


@mock.patch("cg.store.Store")
@mock.patch("cg.apps.vogue.VogueAPI")
@mock.patch("cg.apps.gt.GenotypeAPI")
@mock.patch("cg.store.models.Analysis")
def test_update_analysis_uploaded_to_vogue_date_now(
    mock_analysis, mock_genotype_api, mock_vogue_api, mock_store, timestamp_now
):
    """tests updating the uploaded_to_vogue field of a record in the analysis table"""

    # GIVEN an analysis object with no uploaded_to_vogue date

    # WHEN setting the uploaded to vogue date without specifying a date
    with mock.patch.object(
        UploadVogueAPI.update_analysis_uploaded_to_vogue_date,
        "__defaults__",
        (timestamp_now,),
    ):
        result = UploadVogueAPI(
            mock_genotype_api, mock_vogue_api, mock_store
        ).update_analysis_uploaded_to_vogue_date(mock_analysis)

    # THEN the analysis object should have a vogue_uploaded_date set to the default value
    # dt.datetime.now()
    assert result.uploaded_to_vogue_at == timestamp_now



"""This script tests the cli methods to add families to status-db"""

from cg.cli.add import add
from cg.store import Store
from click.testing import CliRunner
from tests.store_helpers import StoreHelpers
from cg.models.cg_config import CGConfig


def test_add_relationship_required(cli_runner: CliRunner, base_context: CGConfig, helpers):
    """Test to add a relationship using only the required arguments"""
    # GIVEN a database with a sample and an case
    disk_store: Store = base_context.status_db
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    case = helpers.add_case(disk_store)
    case_id = case.internal_id
    status = "affected"

    # WHEN adding a relationship
    result = cli_runner.invoke(
        add, ["relationship", case_id, sample_id, "-s", status], obj=base_context
    )

    # THEN it should be added
    assert result.exit_code == 0
    assert disk_store.FamilySample.query.count() == 1
    assert disk_store.FamilySample.query.first().family.internal_id == case_id
    assert disk_store.FamilySample.query.first().sample.internal_id == sample_id


def test_add_relationship_bad_sample(cli_runner: CliRunner, base_context: CGConfig, helpers):
    """Test to add a relationship using a non-existing sample"""
    # GIVEN an empty database
    disk_store: Store = base_context.status_db
    # WHEN adding a relationship
    case = helpers.add_case(disk_store)
    case_id = case.internal_id
    sample_id = "dummy_sample"
    status = "affected"
    result = cli_runner.invoke(
        add,
        [
            "relationship",
            case_id,
            sample_id,
            "-s",
            status,
        ],
        obj=base_context,
    )

    # THEN it should complain on missing sample instead of adding a relationship
    assert result.exit_code == 1
    assert disk_store.FamilySample.query.count() == 0


def test_add_relationship_bad_family(cli_runner: CliRunner, base_context: CGConfig, helpers):
    """Test to add a relationship using a non-existing case"""
    # GIVEN a database with a sample
    disk_store: Store = base_context.status_db
    # WHEN adding a relationship
    case_id = "dummy_family"
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id

    status = "affected"
    result = cli_runner.invoke(
        add,
        [
            "relationship",
            case_id,
            sample_id,
            "-s",
            status,
        ],
        obj=base_context,
    )

    # THEN it should complain in missing case instead of adding a relationship
    assert result.exit_code == 1
    assert disk_store.FamilySample.query.count() == 0


def test_add_relationship_bad_status(cli_runner: CliRunner, base_context: CGConfig, helpers):
    """Test that the added relationship get the status we send in"""
    # GIVEN a database with a sample and an case
    disk_store: Store = base_context.status_db
    # WHEN adding a relationship

    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    case = helpers.add_case(disk_store)
    case_id = case.internal_id
    status = "dummy_status"

    result = cli_runner.invoke(
        add,
        [
            "relationship",
            case_id,
            sample_id,
            "-s",
            status,
        ],
        obj=base_context,
    )

    # THEN it should complain on bad status instead of adding a relationship
    assert result.exit_code == 2
    assert disk_store.FamilySample.query.count() == 0


def test_add_relationship_mother(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test to add a relationship with a mother"""
    # GIVEN a database with a sample and an case
    disk_store: Store = base_context.status_db
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    mother = helpers.add_sample(disk_store, name="mother")
    mother_id = mother.internal_id
    case = helpers.add_case(disk_store)
    case_id = case.internal_id
    status = "affected"

    # WHEN adding a relationship
    result = cli_runner.invoke(
        add,
        [
            "relationship",
            case_id,
            sample_id,
            "-s",
            status,
            "-m",
            mother_id,
        ],
        obj=base_context,
    )

    # THEN it should be added
    assert result.exit_code == 0
    assert disk_store.FamilySample.query.count() == 1
    assert disk_store.FamilySample.query.first().mother.internal_id == mother_id


def test_add_relationship_bad_mother(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test to add a relationship using a non-existing mother"""
    # GIVEN a database with a sample and an case
    disk_store: Store = base_context.status_db
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    mother_id = "dummy_mother"
    case = helpers.add_case(disk_store)
    case_id = case.internal_id
    status = "affected"

    # WHEN adding a relationship
    result = cli_runner.invoke(
        add,
        [
            "relationship",
            case_id,
            sample_id,
            "-s",
            status,
            "-m",
            mother_id,
        ],
        obj=base_context,
    )

    # THEN it should not be added
    assert result.exit_code == 1
    assert disk_store.FamilySample.query.count() == 0


def test_add_relationship_father(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test to add a relationship using a father"""
    # GIVEN a database with a sample and an case
    disk_store: Store = base_context.status_db
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id

    father = helpers.add_sample(disk_store, name="father", gender="male")
    father_id = father.internal_id

    case = helpers.add_case(disk_store)
    case_id = case.internal_id

    status = "affected"

    # WHEN adding a relationship
    result = cli_runner.invoke(
        add,
        [
            "relationship",
            case_id,
            sample_id,
            "-s",
            status,
            "-f",
            father_id,
        ],
        obj=base_context,
    )

    # THEN it should be added
    assert result.exit_code == 0
    assert disk_store.FamilySample.query.count() == 1
    assert disk_store.FamilySample.query.first().father.internal_id == father_id


def test_add_relationship_bad_father(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test to add a relationship using a non-existing father"""
    # GIVEN a database with a sample and an case
    disk_store: Store = base_context.status_db
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id

    father_id = "bad_father"
    case = helpers.add_case(disk_store)
    case_id = case.internal_id

    status = "affected"

    # WHEN adding a relationship
    result = cli_runner.invoke(
        add,
        [
            "relationship",
            case_id,
            sample_id,
            "-s",
            status,
            "-f",
            father_id,
        ],
        obj=base_context,
    )

    # THEN it should not be added
    assert result.exit_code == 1
    assert disk_store.FamilySample.query.count() == 0



from typing import Optional, List
from sqlalchemy.orm import Query
from cg.store import Store
from cg.store.models import Bed, BedVersion, Customer, Collaboration, Organism, User, Application
from cg.constants.constants import MicrosaltAppTags


def test_get_active_beds(base_store: Store):
    """Test returning not archived bed records from the database."""

    # GIVEN a store with beds

    # WHEN fetching beds
    beds: Query = base_store.get_active_beds()

    # THEN beds should have be returned
    assert beds

    # THEN the bed records should not be archived
    for bed in beds:
        assert bed.is_archived is False


def test_get_active_beds_when_archived(base_store: Store):
    """Test returning not archived bed records from the database when archived."""

    # GIVEN a store with beds
    beds: Query = base_store.get_active_beds()
    for bed in beds:
        bed.is_archived = True
        base_store.add_commit(bed)

    # WHEN fetching beds
    active_beds: Query = base_store.get_active_beds()

    # THEN return no beds
    assert not list(active_beds)


def test_get_bed_by_name(base_store: Store, bed_name: str):
    """Test returning a bed record by name from the database."""

    # GIVEN a store with beds

    # WHEN fetching beds
    bed: Optional[Bed] = base_store.get_bed_by_name(bed_name=bed_name)

    # THEN return a bed
    assert bed

    # THEN return a bed with the supplied bed name
    assert bed.name == bed_name


def test_get_bed_by_name_when_no_match(base_store: Store):
    """Test returning a bed record by name from the database when no match."""

    # GIVEN a store with beds

    # WHEN fetching beds
    bed: Optional[Bed] = base_store.get_bed_by_name(bed_name="does_not_exist")

    # THEN do not return a bed
    assert not bed


def test_get_latest_bed_version(base_store: Store, bed_name: str):
    """Test returning a bed version by bed name from the database."""

    # GIVEN a store with beds

    # WHEN fetching beds
    bed_version: BedVersion = base_store.get_latest_bed_version(bed_name=bed_name)

    # THEN return a bed version with the supplied bed name
    assert bed_version.version == 1


def test_get_application_by_tag(microbial_store: Store, tag: str = MicrosaltAppTags.MWRNXTR003):
    """Test function to return the application by tag."""

    # GIVEN a store with application records

    # WHEN getting the query for the flow cells
    application: Application = microbial_store.get_application_by_tag(tag=tag)

    # THEN return a application with the supplied application tag
    assert application.tag == tag


def test_get_applications_is_not_archived(
    microbial_store: Store, EXPECTED_NUMBER_OF_NOT_ARCHIVED_APPLICATIONS
):
    """Test function to return the application when not archived."""

    # GIVEN a store with application records

    # WHEN getting the query for the flow cells
    applications: List[Application] = microbial_store.get_applications_is_not_archived()

    # THEN return a application with the supplied application tag
    assert len(applications) == EXPECTED_NUMBER_OF_NOT_ARCHIVED_APPLICATIONS
    assert (application.is_archived is False for application in applications)


def test_get_applications_by_prep_category(
    microbial_store: Store,
    EXPECTED_NUMBER_OF_APPLICATIONS_WITH_PREP_CATEGORY,
    prep_category=MicrosaltAppTags.PREP_CATEGORY,
):
    """Test function to return the application by prep category."""

    # GIVEN a store with application records

    # WHEN getting the query for the flow cells
    applications: List[Application] = microbial_store.get_applications_by_prep_category(
        prep_category=prep_category
    )

    # THEN return a application with the supplied application tag
    assert len(applications) == EXPECTED_NUMBER_OF_APPLICATIONS_WITH_PREP_CATEGORY
    assert (application.prep_category == prep_category for application in applications)


def test_get_applications(microbial_store: Store, EXPECTED_NUMBER_OF_APPLICATIONS):
    """Test function to return the applications."""

    # GIVEN a store with application records

    # WHEN getting the query for the flow cells
    applications: List[Application] = microbial_store.get_applications()

    # THEN return a application with the supplied application tag
    assert len(applications) == EXPECTED_NUMBER_OF_APPLICATIONS


def test_get_applications_by_prep_category_and_is_not_archived(
    microbial_store: Store,
    EXPECTED_NUMBER_OF_NOT_ARCHIVED_APPLICATIONS,
    prep_category=MicrosaltAppTags.PREP_CATEGORY,
):
    """Test function to return the application by prep category and not archived."""

    # GIVEN a store with application records

    # WHEN getting the query for the flow cells
    applications: List[
        Application
    ] = microbial_store.get_applications_by_prep_category_and_is_not_archived(
        prep_category=prep_category
    )

    # THEN return a application with the supplied application tag
    assert len(applications) == EXPECTED_NUMBER_OF_NOT_ARCHIVED_APPLICATIONS
    assert (application.prep_category == prep_category for application in applications)
    assert (application.is_archived is False for application in applications)


def test_get_bed_version_query(base_store: Store):
    """Test function to return the bed version query."""

    # GIVEN a store with bed versions records

    # WHEN getting the query for the bed versions
    bed_version_query: Query = base_store._get_query(table=BedVersion)

    # THEN a query should be returned
    assert isinstance(bed_version_query, Query)


def test_get_bed_version_by_short_name(base_store: Store, bed_version_short_name: str):
    """Test function to return the bed version by short name."""

    # GIVEN a store with bed versions records

    # WHEN getting the query for the bed versions
    bed_version: BedVersion = base_store.get_bed_version_by_short_name(
        bed_version_short_name=bed_version_short_name
    )

    # THEN return a bed version with the supplied bed version short name
    assert bed_version.shortname == bed_version_short_name


def test_get_customer_by_customer_id(base_store: Store, customer_id: str):
    """Test function to return the customer by customer id."""

    # GIVEN a store with customer records

    # WHEN getting the query for the customer
    customer: Customer = base_store.get_customer_by_customer_id(customer_id=customer_id)

    # THEN return a customer with the supplied customer internal id
    assert customer.internal_id == customer_id


def test_get_customers(base_store: Store, customer_id: str):
    """Test function to return customers."""

    # GIVEN a store with customer records

    # WHEN getting the customers
    customers: List[Customer] = base_store.get_customers()

    # THEN return customers
    assert customers

    # THEN return a customer with the database customer internal id
    assert customers[0].internal_id == customer_id


def test_get_collaboration_by_internal_id(base_store: Store, collaboration_id: str):
    """Test function to return the collaborations by internal_id."""

    # GIVEN a store with collaborations

    # WHEN getting the query for the collaborations
    collaboration: Collaboration = base_store.get_collaboration_by_internal_id(
        internal_id=collaboration_id
    )

    # THEN return a collaboration with the give collaboration internal_id
    assert collaboration.internal_id == collaboration_id


def test_get_organism_by_internal_id_returns_correct_organism(store_with_organisms: Store):
    """Test finding an organism by internal ID when the ID exists."""

    # GIVEN a store with multiple organisms
    organisms: Query = store_with_organisms._get_query(table=Organism)
    assert organisms.count() > 0

    # GIVEN a random organism from the store
    organism: Organism = organisms.first()
    assert isinstance(organism, Organism)

    # WHEN finding the organism by internal ID
    filtered_organism = store_with_organisms.get_organism_by_internal_id(organism.internal_id)

    # THEN the filtered organism should be of instance Organism
    assert isinstance(filtered_organism, Organism)

    # THEN the filtered organism should match the database organism internal id
    assert filtered_organism.internal_id == organism.internal_id


def test_get_organism_by_internal_id_returns_none_when_id_does_not_exist(
    store_with_organisms: Store,
    non_existent_id: str,
):
    """Test finding an organism by internal ID when the ID does not exist."""

    # GIVEN a store with multiple organisms
    organisms: Query = store_with_organisms._get_query(table=Organism)
    assert organisms.count() > 0

    # WHEN finding the organism by internal ID that does not exist
    filtered_organism: Organism = store_with_organisms.get_organism_by_internal_id(
        internal_id=non_existent_id
    )

    # THEN the filtered organism should be None
    assert filtered_organism is None


def test_get_organism_by_internal_id_returns_none_when_id_is_none(
    store_with_organisms: Store,
):
    """Test finding an organism by internal ID None returns None."""

    # GIVEN a store with multiple organisms
    organisms: Query = store_with_organisms._get_query(table=Organism)
    assert organisms.count() > 0

    # WHEN finding the organism by internal ID None
    filtered_organism: Organism = store_with_organisms.get_organism_by_internal_id(internal_id=None)

    # THEN the filtered organism should be None
    assert filtered_organism is None


def test_get_user_by_email_returns_correct_user(store_with_users: Store):
    """Test fetching a user by email."""

    # GIVEN a store with multiple users
    num_users: int = store_with_users._get_query(table=User).count()
    assert num_users > 0

    # Select a random user from the store
    user: User = store_with_users._get_query(table=User).first()
    assert user is not None

    # WHEN fetching the user by email
    filtered_user: User = store_with_users.get_user_by_email(email=user.email)

    # THEN a user should be returned
    assert isinstance(filtered_user, User)

    # THEN the email should match
    assert filtered_user.email == user.email


def test_get_user_by_email_returns_none_for_nonexisting_email(
    store_with_users: Store, non_existent_email: str
):
    """Test getting user by email when the email does not exist."""

    # GIVEN a non-existing email

    # WHEN retrieving the user by email
    filtered_user: User = store_with_users.get_user_by_email(email=non_existent_email)

    # THEN no user should be returned
    assert filtered_user is None


def test_get_user_when_email_is_none_returns_none(store_with_users: Store):
    """Test getting user by email when the email is None."""

    # WHEN retrieving filtering by email None
    filtered_user: User = store_with_users.get_user_by_email(email=None)

    # THEN no user should be returned
    assert filtered_user is None



"""Module to create delivery reports"""

from datetime import datetime
import logging
from pathlib import Path
from typing import TextIO, Optional, List

import requests
from alchy import Query

from cgmodels.cg.constants import Pipeline
from housekeeper.store.models import File, Version

from cg.constants.constants import FileFormat, MAX_ITEMS_TO_RETRIEVE
from cg.exc import DeliveryReportError
from cg.io.controller import WriteStream
from cg.meta.report.field_validators import get_missing_report_data, get_empty_report_data
from cg.meta.workflow.analysis import AnalysisAPI
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG
from cg.models.analysis import AnalysisModel
from cg.models.cg_config import CGConfig
from cg.meta.meta import MetaAPI
from cg.models.report.metadata import SampleMetadataModel
from cg.models.report.report import (
    ReportModel,
    CustomerModel,
    CaseModel,
    DataAnalysisModel,
    ScoutReportFiles,
)
from cg.models.report.sample import SampleModel, ApplicationModel, TimestampModel, MethodsModel
from cg.store.models import Analysis, Application, Family, Sample
from jinja2 import Environment, PackageLoader, select_autoescape

LOG = logging.getLogger(__name__)


class ReportAPI(MetaAPI):
    """Common Delivery Report API."""

    def __init__(self, config: CGConfig, analysis_api: AnalysisAPI):
        super().__init__(config=config)
        self.analysis_api = analysis_api

    def create_delivery_report(
        self, case_id: str, analysis_date: datetime, force_report: bool
    ) -> str:
        """Generates the html contents of a delivery report."""

        report_data: ReportModel = self.get_report_data(
            case_id=case_id, analysis_date=analysis_date
        )
        report_data: ReportModel = self.validate_report_fields(case_id, report_data, force_report)

        rendered_report = self.render_delivery_report(report_data.dict())
        return rendered_report

    def create_delivery_report_file(
        self, case_id: str, file_path: Path, analysis_date: datetime, force_report: bool
    ) -> TextIO:
        """Generates a temporary file containing a delivery report."""

        file_path.mkdir(parents=True, exist_ok=True)
        delivery_report = self.create_delivery_report(
            case_id=case_id, analysis_date=analysis_date, force_report=force_report
        )

        report_file_path = Path(file_path / "delivery-report.html")
        with open(report_file_path, "w") as delivery_report_file:
            delivery_report_file.write(delivery_report)

        return delivery_report_file

    def add_delivery_report_to_hk(
        self, delivery_report_file: Path, case_id: str, analysis_date: datetime
    ) -> Optional[File]:
        """
        Adds a delivery report file, if it has not already been generated, to an analysis bundle for a specific case
        in HK and returns a pointer to it.
        """

        version = self.housekeeper_api.version(case_id, analysis_date)
        try:
            self.get_delivery_report_from_hk(case_id=case_id)
        except FileNotFoundError:
            LOG.info(f"Adding a new delivery report to housekeeper for {case_id}")
            file: File = self.housekeeper_api.add_file(
                delivery_report_file.name, version, [case_id, HK_DELIVERY_REPORT_TAG]
            )
            self.housekeeper_api.include_file(file, version)
            self.housekeeper_api.add_commit(file)
            return file

        return None

    def get_delivery_report_from_hk(self, case_id: str) -> str:
        """Extracts the delivery reports of a specific case stored in HK."""

        version: Version = self.housekeeper_api.last_version(case_id)
        delivery_report_files: Query = self.housekeeper_api.get_files(
            bundle=case_id, tags=[HK_DELIVERY_REPORT_TAG], version=version.id
        )

        if delivery_report_files.count() == 0:
            LOG.warning(f"No existing delivery report found in housekeeper for {case_id}")
            raise FileNotFoundError

        return delivery_report_files[0].full_path

    def get_scout_uploaded_file_from_hk(self, case_id: str, scout_tag: str) -> Optional[str]:
        """Returns the file path of the uploaded to Scout file given its tag."""

        version: Version = self.housekeeper_api.last_version(bundle=case_id)
        tags: list = self.get_hk_scout_file_tags(scout_tag)
        uploaded_files: Query = self.housekeeper_api.get_files(
            bundle=case_id, tags=tags, version=version.id
        )

        if not tags or uploaded_files.count() == 0:
            LOG.info(
                f"No files were found for the following Scout Housekeeper tag: {scout_tag} (case: {case_id})"
            )

            return None

        return uploaded_files[0].full_path

    def render_delivery_report(self, report_data: dict) -> str:
        """Renders the report on the Jinja template."""

        env = Environment(
            loader=PackageLoader("cg", "meta/report/templates"),
            autoescape=select_autoescape(["html", "xml"]),
        )
        template_name = self.get_template_name()
        template = env.get_template(template_name)
        return template.render(**report_data)

    def get_cases_without_delivery_report(self, pipeline: Pipeline) -> List[Family]:
        """Returns a list of cases that has been stored and need a delivery report."""

        stored_cases = []
        analyses: Query = self.status_db.analyses_to_delivery_report(pipeline)[
            :MAX_ITEMS_TO_RETRIEVE
        ]

        for analysis_obj in analyses:
            case: Family = analysis_obj.family
            last_version: Version = self.housekeeper_api.last_version(case.internal_id)
            hk_file: File = self.housekeeper_api.get_files(
                bundle=case.internal_id, version=last_version.id if last_version else None
            ).first()

            if hk_file and Path(hk_file.full_path).is_file():
                stored_cases.append(case)
            else:
                LOG.warning(
                    f"Case {case.internal_id} must be stored before creating a delivery report"
                )

        return stored_cases

    def get_cases_without_uploaded_delivery_report(self, pipeline: Pipeline) -> List[Family]:
        """Returns a list of cases that need a delivery report to be uploaded."""

        analyses: Query = self.status_db.analyses_to_upload_delivery_reports(pipeline)[
            :MAX_ITEMS_TO_RETRIEVE
        ]

        return [analysis_obj.family for analysis_obj in analyses]

    def update_delivery_report_date(self, case: Family, analysis_date: datetime) -> None:
        """Updates the date when delivery report was created."""

        analysis: Analysis = self.status_db.analysis(case, analysis_date)
        analysis.delivery_report_created_at = datetime.now()
        self.status_db.commit()

    def get_report_data(self, case_id: str, analysis_date: datetime) -> ReportModel:
        """Fetches all the data needed to generate a delivery report."""

        case: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        analysis: Analysis = self.status_db.analysis(case, analysis_date)
        analysis_metadata: AnalysisModel = self.analysis_api.get_latest_metadata(case.internal_id)
        case_model: CaseModel = self.get_case_data(case, analysis, analysis_metadata)

        return ReportModel(
            customer=self.get_customer_data(case),
            version=self.get_report_version(analysis),
            date=datetime.today(),
            case=case_model,
            accredited=self.get_report_accreditation(case_model.samples, analysis_metadata),
        )

    def validate_report_fields(
        self, case_id: str, report_data: ReportModel, force_report
    ) -> ReportModel:
        """Verifies that the required report fields are not empty."""

        required_fields: dict = self.get_required_fields(report_data.case)
        empty_report_fields: dict = get_empty_report_data(report_data)
        missing_report_fields: dict = get_missing_report_data(empty_report_fields, required_fields)

        if missing_report_fields and not force_report:
            LOG.error(
                f"Could not generate report data for {case_id}. "
                f"Missing data: \n{WriteStream.write_stream_from_content(content=missing_report_fields, file_format=FileFormat.YAML)}"
            )
            raise DeliveryReportError

        if empty_report_fields:
            LOG.warning(
                f"Empty report fields: \n{WriteStream.write_stream_from_content(content=empty_report_fields, file_format=FileFormat.YAML)}"
            )

        return report_data

    @staticmethod
    def get_customer_data(case: Family) -> CustomerModel:
        """Returns customer validated attributes retrieved from status DB."""

        return CustomerModel(
            name=case.customer.name,
            id=case.customer.internal_id,
            invoice_address=case.customer.invoice_address,
            scout_access=case.customer.scout_access,
        )

    @staticmethod
    def get_report_version(analysis: Analysis) -> int:
        """
        Returns the version of the given analysis. The version of the first analysis is 1 and subsequent reruns
        increase it by 1.
        """

        version = None

        if analysis:
            version = len(analysis.family.analyses) - analysis.family.analyses.index(analysis)

        return version

    def get_case_data(
        self,
        case: Family,
        analysis: Analysis,
        analysis_metadata: AnalysisModel,
    ) -> CaseModel:
        """Returns case associated validated attributes."""

        samples: List[SampleModel] = self.get_samples_data(case, analysis_metadata)
        unique_applications: List[ApplicationModel] = self.get_unique_applications(samples)

        return CaseModel(
            name=case.name,
            id=case.internal_id,
            data_analysis=self.get_case_analysis_data(case, analysis, analysis_metadata),
            samples=samples,
            applications=unique_applications,
        )

    def get_samples_data(self, case: Family, analysis_metadata: AnalysisModel) -> List[SampleModel]:
        """Extracts all the samples associated to a specific case and their attributes."""

        samples = list()
        case_samples: List[FamilySample] = self.status_db.family_samples(case.internal_id)

        for case_sample in case_samples:
            sample: Sample = case_sample.sample
            lims_sample: Optional[dict] = self.get_lims_sample(sample.internal_id)

            samples.append(
                SampleModel(
                    name=sample.name,
                    id=sample.internal_id,
                    ticket=sample.original_ticket,
                    gender=sample.sex,
                    source=lims_sample.get("source") if lims_sample else None,
                    tumour=sample.is_tumour,
                    application=self.get_sample_application_data(lims_sample),
                    methods=self.get_sample_methods_data(sample.internal_id),
                    status=case_sample.status,
                    metadata=self.get_sample_metadata(case, sample, analysis_metadata),
                    timestamps=self.get_sample_timestamp_data(sample),
                )
            )

        return samples

    def get_lims_sample(self, sample_id: str) -> Optional[dict]:
        """Fetches sample data from LIMS. Returns an empty dictionary if the request was unsuccessful."""

        lims_sample = dict()
        try:
            lims_sample = self.lims_api.sample(sample_id)
        except requests.exceptions.HTTPError as ex:
            LOG.info("Could not fetch sample %s from LIMS: %s", sample_id, ex)

        return lims_sample

    def get_sample_application_data(self, lims_sample: dict) -> ApplicationModel:
        """Retrieves the analysis application attributes."""

        application: Application = self.status_db.get_application_by_tag(
            tag=lims_sample.get("application")
        )

        return (
            ApplicationModel(
                tag=application.tag,
                version=lims_sample.get("application_version"),
                prep_category=application.prep_category,
                description=application.description,
                limitations=application.limitations,
                accredited=application.is_accredited,
                external=application.is_external,
            )
            if application
            else ApplicationModel()
        )

    @staticmethod
    def get_unique_applications(samples: List[SampleModel]) -> List[ApplicationModel]:
        """Returns the unique case associated applications."""

        applications = list()
        for sample in samples:
            if sample.application not in applications:
                applications.append(sample.application)

        return applications

    def get_sample_methods_data(self, sample_id: str) -> MethodsModel:
        """Fetches sample library preparation and sequencing methods from LIMS."""

        library_prep = None
        sequencing = None
        try:
            library_prep = self.lims_api.get_prep_method(sample_id)
            sequencing = self.lims_api.get_sequencing_method(sample_id)
        except requests.exceptions.HTTPError as ex:
            LOG.info("Could not fetch sample (%s) methods from LIMS: %s", sample_id, ex)

        return MethodsModel(library_prep=library_prep, sequencing=sequencing)

    def get_case_analysis_data(
        self,
        case: Family,
        analysis: Analysis,
        analysis_metadata: AnalysisModel,
    ) -> DataAnalysisModel:
        """Retrieves the pipeline attributes used for data analysis."""

        return DataAnalysisModel(
            customer_pipeline=case.data_analysis,
            data_delivery=case.data_delivery,
            pipeline=analysis.pipeline,
            pipeline_version=analysis.pipeline_version,
            type=self.get_data_analysis_type(case),
            genome_build=self.get_genome_build(analysis_metadata),
            variant_callers=self.get_variant_callers(analysis_metadata),
            panels=case.panels,
            scout_files=self.get_scout_uploaded_files(case),
        )

    def get_scout_uploaded_files(self, case: Family) -> ScoutReportFiles:
        """Extracts the files that will be uploaded to Scout."""

        return ScoutReportFiles(
            snv_vcf=self.get_scout_uploaded_file_from_hk(case.internal_id, "snv_vcf"),
            sv_vcf=self.get_scout_uploaded_file_from_hk(case.internal_id, "sv_vcf"),
            vcf_str=self.get_scout_uploaded_file_from_hk(case.internal_id, "vcf_str"),
            smn_tsv=self.get_scout_uploaded_file_from_hk(case.internal_id, "smn_tsv"),
        )

    @staticmethod
    def get_sample_timestamp_data(sample: Sample) -> TimestampModel:
        """Retrieves the sample processing dates."""

        return TimestampModel(
            ordered_at=sample.ordered_at,
            received_at=sample.received_at,
            prepared_at=sample.prepared_at,
            sequenced_at=sample.sequenced_at,
        )

    def get_sample_metadata(
        self,
        case: Family,
        sample: Sample,
        analysis_metadata: AnalysisModel,
    ) -> SampleMetadataModel:
        """Fetches the sample metadata to include in the report."""

        raise NotImplementedError

    def get_data_analysis_type(self, case: Family) -> Optional[str]:
        """Retrieves the data analysis type carried out."""

        raise NotImplementedError

    def get_genome_build(self, analysis_metadata: AnalysisModel) -> str:
        """Returns the build version of the genome reference of a specific case."""

        raise NotImplementedError

    def get_variant_callers(self, analysis_metadata: AnalysisModel) -> list:
        """Extracts the list of variant-calling filters used during analysis."""

        raise NotImplementedError

    def get_report_accreditation(
        self, samples: List[SampleModel], analysis_metadata: AnalysisModel
    ) -> bool:
        """Checks if the report is accredited or not."""

        raise NotImplementedError

    def get_required_fields(self, case: CaseModel) -> dict:
        """Retrieves a dictionary with the delivery report required fields."""

        raise NotImplementedError

    def get_template_name(self) -> str:
        """Retrieves the pipeline specific template name."""

        raise NotImplementedError

    @staticmethod
    def get_application_required_fields(case: CaseModel, required_fields: list) -> dict:
        """Retrieves sample required fields."""

        required_sample_fields = dict()

        for application in case.applications:
            required_sample_fields.update({application.tag: required_fields})

        return required_sample_fields

    @staticmethod
    def get_sample_required_fields(case: CaseModel, required_fields: list) -> dict:
        """Retrieves sample required fields."""

        required_sample_fields = dict()

        for sample in case.samples:
            required_sample_fields.update({sample.id: required_fields})

        return required_sample_fields

    @staticmethod
    def get_timestamp_required_fields(case: CaseModel, required_fields: list) -> dict:
        """Retrieves sample timestamps required fields."""

        for sample in case.samples:
            if sample.application.external:
                required_fields.remove("received_at")
                break

        return ReportAPI.get_sample_required_fields(case, required_fields)

    def get_hk_scout_file_tags(self, scout_tag: str) -> Optional[list]:
        """Retrieves pipeline specific uploaded to Scout Housekeeper file tags given a Scout key."""

        tags = self.get_upload_case_tags().get(scout_tag)

        return list(tags) if tags else None

    def get_upload_case_tags(self):
        """Retrieves pipeline specific upload case tags."""

        raise NotImplementedError



from datetime import datetime as dt

from cg.store import Store
from cg.store.models import Customer


def test_add_collaboration(store: Store):
    # GIVEN an empty database
    assert store.Collaboration.query.first() is None
    internal_id, name = "cust_group", "Test customer group"

    # WHEN adding a new customer group
    new_collaboration = store.add_collaboration(internal_id=internal_id, name=name)
    store.add_commit(new_collaboration)

    # THEN it should be stored in the database
    assert store.Collaboration.query.first() == new_collaboration


def test_add_user(store: Store):
    # GIVEN a database with a customer in it that we can connect the user to
    customer = store.add_customer(
        internal_id="custtest",
        name="Test Customer",
        scout_access=False,
        invoice_address="dummy street 1",
        invoice_reference="dummy nr",
    )
    store.add_commit(customer)

    # WHEN adding a new user
    name, email = "Paul T. Anderson", "paul.anderson@magnolia.com"
    new_user = store.add_user(customer=customer, email=email, name=name)

    store.add_commit(new_user)

    # THEN it should be stored in the database
    assert store.User.query.first() == new_user


def test_add_microbial_sample(base_store: Store, helpers):
    # GIVEN an empty database
    assert base_store.Sample.query.first() is None
    customer_obj = helpers.ensure_customer(base_store)
    assert customer_obj
    name = "microbial_sample"
    organism_name = "e. coli"
    internal_id = "lims-id"
    reference_genome = "ref_gen"
    priority = "research"
    application_version = base_store.ApplicationVersion.query.first()
    base_store.add_organism(organism_name, organism_name, reference_genome)
    organism = base_store.Organism.query.first()

    # WHEN adding a new microbial sample
    new_sample = base_store.add_sample(
        name=name,
        sex="unknown",
        internal_id=internal_id,
        priority=priority,
        application_version=application_version,
        organism=organism,
        reference_genome=reference_genome,
    )
    new_sample.customer = customer_obj
    base_store.add_commit(new_sample)

    # THEN it should be stored in the database
    assert base_store.Sample.query.first() == new_sample
    stored_microbial_sample = base_store.Sample.query.first()
    assert stored_microbial_sample.name == name
    assert stored_microbial_sample.internal_id == internal_id
    assert stored_microbial_sample.reference_genome == reference_genome
    assert stored_microbial_sample.application_version == application_version
    assert stored_microbial_sample.priority_human == priority
    assert stored_microbial_sample.organism == organism


def test_add_pool(rml_pool_store: Store):
    """Tests whether new pools are invoiced as default"""
    # GIVEN a valid customer and a valid application_version
    customer: Customer = rml_pool_store.get_customers()[0]
    application = rml_pool_store.get_application_by_tag(tag="RMLP05R800")
    app_version = rml_pool_store.application_version(application=application, version=1)

    # WHEN adding a new pool
    new_pool = rml_pool_store.add_pool(
        customer=customer,
        name="pool2",
        order="123456",
        ordered=dt.now(),
        application_version=app_version,
    )

    rml_pool_store.add_commit(new_pool)
    # THEN the new pool should have no_invoice = False
    pool = rml_pool_store.get_pool_by_entry_id(entry_id=2)
    assert pool.no_invoice is False



from typing import Optional

from sqlalchemy import Column, types
from sqlalchemy.orm.exc import NoResultFound

from .base import Model


class Backuptape(Model):
    backuptape_id = Column(types.Integer, primary_key=True)
    tapedir = Column(types.String(255))
    nametext = Column(types.String(255))
    tapedate = Column(types.DateTime)

    @staticmethod
    def exists(tapedir: str) -> Optional[int]:
        """Check if a tape already exists
        tapedir (str): the name of the tape, e.g. tape036_037
        """
        try:
            backup_tape: Backuptape = Backuptape.query.filter_by(tapedir=tapedir).one()
            return backup_tape.backuptape_id
        except NoResultFound:
            return None



"""Handler to find basic data objects"""
import datetime as dt
from typing import List, Optional

from sqlalchemy import desc
from sqlalchemy.orm import Query

from cg.store.models import (
    Application,
    ApplicationVersion,
    Bed,
    BedVersion,
    Customer,
    Collaboration,
    Organism,
    Panel,
    User,
)
from cg.store.api.base import BaseHandler
from cg.store.filters.status_organism_filters import OrganismFilter, apply_organism_filter
from cg.store.filters.status_bed_filters import apply_bed_filter, BedFilter
from cg.store.filters.status_bed_version_filters import BedVersionFilter, apply_bed_version_filter
from cg.store.filters.status_customer_filters import apply_customer_filter, CustomerFilter
from cg.store.filters.status_application_filters import apply_application_filter, ApplicationFilter
from cg.store.filters.status_collaboration_filters import (
    CollaborationFilter,
    apply_collaboration_filter,
)
from cg.store.filters.status_panel_filters import PanelFilter, apply_panel_filter
from cg.store.filters.status_user_filters import apply_user_filter, UserFilter


class FindBasicDataHandler(BaseHandler):
    """Contains methods to find basic data model instances."""

    def get_application_by_tag(self, tag: str) -> Application:
        """Return an application by tag."""
        return apply_application_filter(
            applications=self._get_query(table=Application),
            filter_functions=[ApplicationFilter.FILTER_BY_TAG],
            tag=tag,
        ).first()

    def get_applications_by_prep_category(self, prep_category: str) -> List[Application]:
        """Return applications by prep category."""
        return (
            apply_application_filter(
                applications=self._get_query(table=Application),
                filter_functions=[ApplicationFilter.FILTER_BY_PREP_CATEGORY],
                prep_category=prep_category,
            )
            .order_by(self.Application.prep_category, self.Application.tag)
            .all()
        )

    def get_applications_is_not_archived(self) -> List[Application]:
        """Return applications that are not archived."""
        return (
            apply_application_filter(
                applications=self._get_query(table=Application),
                filter_functions=[ApplicationFilter.FILTER_IS_NOT_ARCHIVED],
            )
            .order_by(self.Application.prep_category, self.Application.tag)
            .all()
        )

    def get_applications_by_prep_category_and_is_not_archived(
        self, prep_category: str
    ) -> List[Application]:
        """Return applications by prep category that are not archived."""
        return (
            apply_application_filter(
                applications=self._get_query(table=Application),
                filter_functions=[
                    ApplicationFilter.FILTER_BY_PREP_CATEGORY,
                    ApplicationFilter.FILTER_IS_NOT_ARCHIVED,
                ],
                prep_category=prep_category,
            )
            .order_by(self.Application.prep_category, self.Application.tag)
            .all()
        )

    def get_applications_by_prep_category_and_is_archived(
        self, prep_category: str
    ) -> List[Application]:
        """Return applications by prep category that are archived."""
        return (
            apply_application_filter(
                applications=self._get_query(table=Application),
                filter_functions=[
                    ApplicationFilter.FILTER_BY_PREP_CATEGORY,
                    ApplicationFilter.FILTER_IS_ARCHIVED,
                ],
                prep_category=prep_category,
            )
            .order_by(self.Application.prep_category, self.Application.tag)
            .all()
        )

    def get_applications(self) -> List[Application]:
        """Return all applications."""
        return (
            self._get_query(table=Application)
            .order_by(self.Application.prep_category, self.Application.tag)
            .all()
        )

    def application_version(self, application: Application, version: int) -> ApplicationVersion:
        """Fetch an application version."""
        query = self.ApplicationVersion.query.filter_by(application=application, version=version)
        return query.first()

    def get_bed_version_by_short_name(self, bed_version_short_name: str) -> BedVersion:
        """Return bed version with short name."""
        return apply_bed_version_filter(
            bed_versions=self._get_query(table=BedVersion),
            bed_version_short_name=bed_version_short_name,
            filter_functions=[BedVersionFilter.FILTER_BY_SHORT_NAME],
        ).first()

    def get_bed_by_name(self, bed_name: str) -> Optional[Bed]:
        """Return bed by name."""
        return apply_bed_filter(
            beds=self._get_query(table=Bed),
            bed_name=bed_name,
            filter_functions=[BedFilter.FILTER_BY_NAME],
        ).first()

    def get_active_beds(self) -> Query:
        """Get all beds which are not archived."""
        bed_filter_functions: List[BedFilter] = [
            BedFilter.FILTER_NOT_ARCHIVED,
            BedFilter.ORDER_BY_NAME,
        ]
        return apply_bed_filter(
            beds=self._get_query(table=Bed), filter_functions=bed_filter_functions
        )

    def get_latest_bed_version(self, bed_name: str) -> Optional[BedVersion]:
        """Return the latest bed version for a bed by supplied name."""
        bed: Optional[Bed] = self.get_bed_by_name(bed_name=bed_name)
        return bed.versions[-1] if bed and bed.versions else None

    def get_customer_by_customer_id(self, customer_id: str) -> Customer:
        """Return customer with customer id."""
        return apply_customer_filter(
            filter_functions=[CustomerFilter.FILTER_BY_INTERNAL_ID],
            customers=self._get_query(table=Customer),
            customer_internal_id=customer_id,
        ).first()

    def get_collaboration_by_internal_id(self, internal_id: str) -> Collaboration:
        """Fetch a customer group by internal id from the store."""
        return apply_collaboration_filter(
            collaborations=self._get_query(table=Collaboration),
            filter_functions=[CollaborationFilter.FILTER_BY_INTERNAL_ID],
            internal_id=internal_id,
        ).first()

    def current_application_version(self, tag: str) -> Optional[ApplicationVersion]:
        """Fetch the current application version for an application tag."""
        application_obj = self.Application.query.filter_by(tag=tag).first()
        if not application_obj:
            return None
        application_id = application_obj.id
        records = self.ApplicationVersion.query.filter_by(application_id=application_id)
        records = records.filter(self.ApplicationVersion.valid_from < dt.datetime.now())
        records = records.order_by(desc(self.ApplicationVersion.valid_from))

        return records.first()

    def latest_version(self, tag: str) -> Optional[ApplicationVersion]:
        """Fetch the latest application version for an application tag."""
        application_obj = self.Application.query.filter_by(tag=tag).first()
        return (
            application_obj.versions[-1] if application_obj and application_obj.versions else None
        )

    def get_organism_by_internal_id(self, internal_id: str) -> Organism:
        """Find an organism by internal id."""
        return apply_organism_filter(
            organisms=self._get_query(table=Organism),
            filter_functions=[OrganismFilter.FILTER_BY_INTERNAL_ID],
            internal_id=internal_id,
        ).first()

    def get_all_organisms(self) -> List[Organism]:
        """Return all organisms ordered by organism internal id."""
        return self._get_query(table=Organism).order_by(Organism.internal_id)

    def get_customers(self) -> List[Customer]:
        """Return costumers."""
        return self._get_query(table=Customer).all()

    def get_panel_by_abbreviation(self, abbreviation: str) -> Panel:
        """Return a panel by abbreviation."""
        return apply_panel_filter(
            panels=self._get_query(table=Panel),
            filters=[PanelFilter.FILTER_BY_ABBREVIATION],
            abbreviation=abbreviation,
        ).first()

    def get_panels(self) -> List[Panel]:
        """Returns all panels."""
        return self._get_query(table=Panel).order_by(Panel.abbrev).all()

    def get_user_by_email(self, email: str) -> User:
        """Return a user by email from the database."""
        return apply_user_filter(
            users=self._get_query(table=User),
            email=email,
            filter_functions=[UserFilter.FILTER_BY_EMAIL],
        ).first()



from pathlib import Path
from typing import List

import pytest

from cg.apps.demultiplex.sample_sheet.index import Index
from cg.apps.demultiplex.sample_sheet.novaseq_sample_sheet import SampleSheetCreator
from cg.apps.lims.samplesheet import (
    LimsFlowcellSample,
    LimsFlowcellSampleBcl2Fastq,
    LimsFlowcellSampleDragen,
)
from cg.models.demultiplex.run_parameters import RunParameters


@pytest.fixture(name="output_dirs_bcl2fastq")
def fixture_output_dirs_bcl2fastq(demultiplexed_runs: Path) -> Path:
    """Return the output path a dir with flow cells that have finished demultiplexing using
    bcl2fastq."""
    return Path(demultiplexed_runs, "bcl2fastq")


@pytest.fixture(name="demux_run_dir_bcl2fastq")
def fixture_demux_run_dir_bcl2fastq(demux_run_dir: Path) -> Path:
    """Return the path to a dir with flowcells ready for demultiplexing"""
    return Path(demux_run_dir, "bcl2fastq")


@pytest.fixture(name="demux_run_dir_dragen")
def fixture_demux_run_dir_dragen(demux_run_dir: Path) -> Path:
    """Return the path to a dir with flowcells ready for demultiplexing"""
    return Path(demux_run_dir, "dragen")


@pytest.fixture(name="index_obj")
def fixture_index_obj() -> Index:
    return Index(name="C07 - UDI0051", sequence="AACAGGTT-ATACCAAG")


@pytest.fixture(name="novaseq_dir")
def fixture_novaseq_dir(demux_run_dir: Path, flow_cell_full_name: str) -> Path:
    """Return the path to the novaseq demultiplex fixtures"""
    return Path(demux_run_dir, flow_cell_full_name)


@pytest.fixture(name="flow_cell_dir_bcl2fastq")
def fixture_novaseq_dir_bcl2fastq(demux_run_dir_bcl2fastq: Path, flow_cell_full_name: str) -> Path:
    """Return the path to the novaseq demultiplex fixtures"""
    return Path(demux_run_dir_bcl2fastq, flow_cell_full_name)


@pytest.fixture(name="flow_cell_dir_dragen")
def fixture_novaseq_dir_dragen(demux_run_dir_dragen: Path, flow_cell_full_name: str) -> Path:
    """Return the path to the novaseq demultiplex fixtures"""
    return Path(demux_run_dir_dragen, flow_cell_full_name)


@pytest.fixture(name="hiseq_dir")
def fixture_hiseq_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the novaseq demultiplex fixtures"""
    return Path(demultiplex_fixtures, "hiseq_run")


@pytest.fixture(name="unknown_run_parameters")
def fixture_unknown_run_parameters(demultiplex_fixtures: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return Path(demultiplex_fixtures, "unknown_run_parameters.xml")


@pytest.fixture(name="run_parameters_missing_flowcell_type")
def fixture_run_parameters_missing_flowcell_type(demultiplex_fixtures: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return Path(demultiplex_fixtures, "runParameters_missing_flowcell_run_field.xml")


@pytest.fixture(name="hiseq_run_parameters")
def fixture_hiseq_run_parameters(hiseq_dir: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return Path(hiseq_dir, "runParameters.xml")


@pytest.fixture(name="novaseq_run_parameters")
def fixture_novaseq_run_parameters(novaseq_dir: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return Path(novaseq_dir, "RunParameters.xml")


@pytest.fixture(name="raw_lims_sample")
def fixture_raw_lims_sample(flow_cell_name: str) -> LimsFlowcellSample:
    """Return a raw lims sample"""
    sample = {
        "flowcell_id": flow_cell_name,
        "lane": 1,
        "sample_id": "ACC7628A20",
        "sample_ref": "hg19",
        "index": "ACTGGTGTCG-ACAGGACTTG",
        "description": "",
        "sample_name": "814206",
        "control": "N",
        "recipe": "R1",
        "operator": "script",
        "project": "814206",
    }
    return LimsFlowcellSample(**sample)


@pytest.fixture(name="lims_novaseq_samples")
def fixture_lims_novaseq_samples(lims_novaseq_samples_raw: List[dict]) -> List[LimsFlowcellSample]:
    """Return a list of parsed flowcell samples"""
    return [LimsFlowcellSample(**sample) for sample in lims_novaseq_samples_raw]


@pytest.fixture(name="lims_novaseq_bcl2fastq_samples")
def fixture_lims_novaseq_bcl2fastq_samples(
    lims_novaseq_samples_raw: List[dict],
) -> List[LimsFlowcellSampleBcl2Fastq]:
    """Return a list of parsed flow cell samples"""
    return [LimsFlowcellSampleBcl2Fastq(**sample) for sample in lims_novaseq_samples_raw]


@pytest.fixture(name="lims_novaseq_dragen_samples")
def fixture_lims_novaseq_dragen_samples(
    lims_novaseq_samples_raw: List[dict],
) -> List[LimsFlowcellSampleDragen]:
    """Return a list of parsed flowcell samples"""
    return [LimsFlowcellSampleDragen(**sample) for sample in lims_novaseq_samples_raw]


@pytest.fixture(name="novaseq_run_parameters_object")
def fixture_novaseq_run_parameters_object(novaseq_run_parameters: Path) -> RunParameters:
    return RunParameters(novaseq_run_parameters)


@pytest.fixture(name="novaseq_bcl2fastq_sample_sheet_object")
def fixture_novaseq_bcl2fastq_sample_sheet_object(
    flow_cell_id: str,
    lims_novaseq_bcl2fastq_samples: List[LimsFlowcellSampleBcl2Fastq],
    novaseq_run_parameters_object: RunParameters,
) -> SampleSheetCreator:
    return SampleSheetCreator(
        flowcell_id=flow_cell_id,
        lims_samples=lims_novaseq_bcl2fastq_samples,
        run_parameters=novaseq_run_parameters_object,
        bcl_converter="bcl2fastq",
    )


@pytest.fixture(name="novaseq_dragen_sample_sheet_object")
def fixture_novaseq_dragen_sample_sheet_object(
    flow_cell_id: str,
    lims_novaseq_dragen_samples: List[LimsFlowcellSampleDragen],
    novaseq_run_parameters_object: RunParameters,
) -> SampleSheetCreator:
    return SampleSheetCreator(
        flowcell_id=flow_cell_id,
        lims_samples=lims_novaseq_dragen_samples,
        run_parameters=novaseq_run_parameters_object,
        bcl_converter="dragen",
    )
